import asyncio
import logging
import warnings
from dataclasses import dataclass
from collections import defaultdict
from typing import (
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Literal,
    TypeVar,
    Iterable,
    Optional,
    Sequence,
    Collection,
    cast,
    overload,
)

import numpy as np

from qm.qua import fixed
from qm import DictQuaConfig
from qm.persistence import BaseStore
from qm.grpc.qua_config import QuaConfig
from qm.utils import deprecation_message
from qm.utils.async_utils import run_async
from qm.results import StreamingResultFetcher
from qm.api.v2.job_result_api import JobResultApi
from qm._report import ExecutionError, ExecutionReport
from qm.type_hinting import Value, NumpySupportedValue
from qm.api.v2.job_api.generic_apis import JobGenericApi
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QmValueError, FunctionInputError
from qm.program.ConfigBuilder import convert_msg_to_config
from qm.api.v2.job_api.job_elements_db import JobElementsDB
from qm.utils.general_utils import create_input_stream_name
from qm.grpc.job_manager import IntStreamData, BoolStreamData, FixedStreamData
from qm.api.v2.job_api.element_input_api import MwInputApi, MixInputsApi, SingleInputApi
from qm.grpc.v2 import (
    JobMetadata,
    CancelRequest,
    ResumeRequest,
    JobResponseData,
    GetIoValuesRequest,
    JobExecutionStatus,
    SetIoValuesRequest,
    GetJobErrorsRequest,
    JobServiceIsPausedRequest,
    JobServiceGetConfigRequest,
    JobServiceGetJobStatusRequest,
    SetIoValuesRequestIoValueSetData,
    JobServicePushToInputStreamRequest,
    GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData,
)

logger = logging.getLogger(__name__)

NumberType = Union[bool, int, float]
NumberTypeVar = TypeVar("NumberTypeVar", bool, int, float)
JobStatus = Literal["In queue", "Running", "Processing", "Done", "Canceled", "Error"]

JOB_STATUS_MAPPING: Dict[JobExecutionStatus, JobStatus] = {
    JobExecutionStatus.UNKNOWN: "Error",
    JobExecutionStatus.PENDING: "In queue",
    JobExecutionStatus.RUNNING: "Running",
    JobExecutionStatus.COMPLETED: "Done",
    JobExecutionStatus.CANCELED: "Canceled",
    JobExecutionStatus.LOADING: "In queue",
    JobExecutionStatus.ERROR: "Error",
    JobExecutionStatus.PROCESSING: "Processing",
}
_inverse_job_status_mapping_tmp = defaultdict(list)
for k, v in JOB_STATUS_MAPPING.items():
    _inverse_job_status_mapping_tmp[v].append(k)
_INVERSE_JOB_STATUS_MAPPING = dict(_inverse_job_status_mapping_tmp)

IoValueTypes = Union[Type[bool], Type[int], Type[float], Type[fixed]]


class JobFailedError(ValueError):
    pass


def transfer_statuses_to_enum(status: Union[JobStatus, Iterable[JobStatus]]) -> List[JobExecutionStatus]:
    if isinstance(status, str):
        status = [status]
    try:
        return sum([_INVERSE_JOB_STATUS_MAPPING[x] for x in status], [])
    except KeyError:
        raise QmValueError(f"One ore more statuses is invalid: {status}")


@overload
def _extract_io_value_type(
    io_value: GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, io_type: None
) -> GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData:
    pass


@overload
def _extract_io_value_type(
    io_value: GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, io_type: Type[bool]
) -> bool:
    pass


@overload
def _extract_io_value_type(
    io_value: GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, io_type: Type[int]
) -> int:
    pass


@overload
def _extract_io_value_type(
    io_value: GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, io_type: Union[Type[float], Type[fixed]]
) -> float:
    pass


def _extract_io_value_type(
    io_value: GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, io_type: Optional[IoValueTypes]
) -> Union[bool, int, float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
    if io_type is None:
        return io_value
    if io_type == bool:
        return io_value.boolean_value
    if io_type == int:
        return io_value.int_value
    if io_type == float or io_type == fixed:
        return io_value.double_value
    raise TypeError(f"Data type {io_type} is not supported")


@dataclass
class JobData:
    id: str
    status: JobStatus
    description: str
    metadata: JobMetadata  # This is a GRPC class, consider replacing with a dataclass

    @classmethod
    def from_grpc(cls, grpc_job: JobResponseData) -> "JobData":
        return cls(
            id=grpc_job.job_id,
            status=JOB_STATUS_MAPPING[grpc_job.status],
            description=grpc_job.description,
            metadata=grpc_job.metadata,
        )


class JobApi(JobGenericApi):
    def __init__(
        self, connection_details: ConnectionDetails, job_id: str, store: BaseStore, capabilities: ServerCapabilities
    ) -> None:
        super().__init__(connection_details, job_id)
        self._elements: Optional[JobElementsDB] = None
        self._store = store
        self._caps = capabilities

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(connection_details={self._connection_details}, job_id={self._id})"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}:\n\tid: {self._id}\n\tstatus: {self.get_status()}"

    @property
    def elements(self) -> JobElementsDB:
        if self._elements is None:  # This is done lazily to avoid unnecessary calls to the server
            self._elements = self._create_elements()
        return self._elements

    @property
    def id(self) -> str:
        return self._id

    def get_job_id(self) -> str:
        """
        Returns:
             The job's ID
        """
        return self.id

    def _create_elements(self) -> JobElementsDB:
        config = self._get_pb_config()
        elements_config = config.v1_beta.elements
        return JobElementsDB.init_from_data(elements_config, self._connection_details, self._id)

    def _get_pb_config(self) -> QuaConfig:
        request = JobServiceGetConfigRequest(job_id=self._id)
        return self._run(self._stub.get_config(request, timeout=self._timeout)).config

    def get_compilation_config(self) -> DictQuaConfig:
        """
        Returns:
             The config with which this job was compiled
        """
        return convert_msg_to_config(self._get_pb_config())

    def push_to_input_stream(self, stream_name: str, data: List[Union[bool, int, float]]) -> None:
        """Push data to the input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.

        Multiple data entries can be inserted before the data is read by the program.

        See [Input streams](../Guides/features.md#input-streams) for more information.

        Args:
            stream_name: The input stream name the data is to be inserted to.
            data: The data to be inserted. The data's size must match
                the size of the input stream.
        """
        if all(type(element) == bool for element in data):
            self._typed_push_to_input_stream(stream_name, bool, data)
        elif all(type(element) == int for element in data):
            self._typed_push_to_input_stream(stream_name, int, data)
        elif all(type(element) == float for element in data):
            self._typed_push_to_input_stream(stream_name, float, data)
        else:
            raise QmValueError(
                f"Invalid type in data, type is '{set(type(el) for el in data)}', "
                f"excepted types are bool | int | float"
            )

    def _typed_push_to_input_stream(
        self, stream_name: str, data_type: Type[NumberTypeVar], data: Sequence[NumberTypeVar]
    ) -> None:
        stream_name = create_input_stream_name(stream_name)
        if data_type == bool:
            request = JobServicePushToInputStreamRequest(
                job_id=self._id, stream_name=stream_name, bool_stream_data=BoolStreamData([bool(d) for d in data])
            )
        elif data_type == int:
            request = JobServicePushToInputStreamRequest(
                job_id=self._id, stream_name=stream_name, int_stream_data=IntStreamData([int(d) for d in data])
            )
        elif data_type == float:
            request = JobServicePushToInputStreamRequest(
                job_id=self._id, stream_name=stream_name, fixed_stream_data=FixedStreamData([float(d) for d in data])
            )
        else:
            raise TypeError(f"Data type {data_type} is not supported")

        self._run(self._stub.push_to_input_stream(request, timeout=self._timeout))

    def get_status(self) -> JobStatus:
        """
        Returns the status of the job, one of the following strings:

        - `In queue` - Program is in the queue
        - `Running` - Program is currently running
        - `Processing` - Program done but data processing is ongoing
        - `Done` - Program and data processing done
        - `Canceled` - Program was canceled before it was done
        - `Error` - Program encountered an error

        Returns:

             The job status
        """
        request = JobServiceGetJobStatusRequest(job_id=self._id)
        response = self._run(self._stub.get_job_status(request, timeout=self._timeout))
        return JOB_STATUS_MAPPING[response.status]

    def is_running(self) -> bool:
        """
        Returns:

             `True` if the job is currently running
        """
        return self.get_status() == "Running"

    def wait_until(self, state: Union[JobStatus, Collection[JobStatus]], timeout: float) -> None:
        """
        Waits until a specific state is reached. If the job is already passed the given state, the function will
        immediately return. See [get_status][qm.api.v2.job_api.job_api.JobApi.get_status] for a list of statuses.

        If the state cannot be reached (e.g., waiting for “Done” but the job is “Canceled”) then an error is raised.

        If the timeout time, has passed, an error is raised.

        Args:
            state: The state to wait for
            timeout: The timeout time, in seconds
        """
        if isinstance(state, str):
            state = {state}
        try:
            run_async(self._wait_until(state, timeout))
        except asyncio.exceptions.TimeoutError:
            raise TimeoutError(f"Job {self.id} did not reach any state of {state} within {timeout} seconds")

    async def _wait_until(self, states: Collection[JobStatus], timeout: float) -> None:
        general_timout = self._timeout if self._timeout is not None else float("inf")
        start_time = asyncio.get_event_loop().time()
        request = JobServiceGetJobStatusRequest(job_id=self._id)
        _timeout = min(timeout, general_timout)
        async for status in self._stub.get_job_status_updates(request, timeout=_timeout):
            # The timout above is a bit an abuse, it is the timeout of the request and not of the status polling
            status_str = JOB_STATUS_MAPPING[status.success.status]
            if status_str in states:
                logger.debug(f"Job {self.id} reached state {status_str}")
                return
            if status_str == "Done":
                logger.info(f"Job {self.id} is done")
                return
            if status_str in {"Error", "Canceled"}:
                raise JobFailedError(f"Job {self.id} reached state {status_str}")

            elapsed_time = asyncio.get_event_loop().time() - start_time
            time_remained = timeout - elapsed_time
            if time_remained < 0:
                raise TimeoutError(f"Job {self.id} did not reach any state of {states} within {timeout} seconds")

            _timeout = min(time_remained, general_timout)
            await asyncio.sleep(0.1)  # I (YR) just put a number here that sounds reasonable

    def is_finished(self) -> bool:
        """
        Returns:

             `True` if the job will no longer run (Has reached "Completed", "Canceled" or "Error").
        """
        return self.get_status() in {"Completed", "Canceled", "Error"}

    def cancel(self) -> None:
        """
        Cancels the job
        """
        request = CancelRequest(job_id=self._id)
        self._run(self._stub.cancel(request, timeout=self._timeout))

    def is_paused(self) -> bool:
        """
        Returns:

             `True` if the job was paused from QUA.
        """
        request = JobServiceIsPausedRequest(job_id=self._id)
        response = self._run(self._stub.is_paused(request, timeout=self._timeout))
        return response.is_paused

    def resume(self) -> None:
        """Resumes a program that was halted using the [pause][qm.qua._dsl.pause] statement"""
        request = ResumeRequest(job_id=self._id)
        self._run(self._stub.resume(request, timeout=self._timeout))

    @staticmethod
    def _fix_io_value_type(value: Optional[NumpySupportedValue]) -> SetIoValuesRequestIoValueSetData:
        if value is None:
            return SetIoValuesRequestIoValueSetData()
        if isinstance(value, (np.bool_, bool)):
            return SetIoValuesRequestIoValueSetData(boolean_value=bool(value))
        if isinstance(value, (np.integer, int)):
            return SetIoValuesRequestIoValueSetData(int_value=int(value))
        if isinstance(value, (np.floating, float)):
            return SetIoValuesRequestIoValueSetData(double_value=float(value))
        raise QmValueError(f"cannot convert {type(value)} to int | float | bool")

    def set_io_values(
        self, io1: Optional[NumpySupportedValue] = None, io2: Optional[NumpySupportedValue] = None
    ) -> None:
        """
        Sets the values of ``IO1`` & ``IO2`. If only one is given, then the value of the second will remain unchanged.

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            io1: The value to be placed in ``IO1``
            io2: The value to be placed in ``IO2``
        """
        request = SetIoValuesRequest(
            job_id=self._id,
            io1=self._fix_io_value_type(io1),
            io2=self._fix_io_value_type(io2),
        )
        self._run(self._stub.set_io_values(request, timeout=self._timeout))

    def set_io1_value(self, value: Optional[NumpySupportedValue]) -> None:
        """
        Sets the values of ``IO1``.

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value: The value to be placed in ``IO1``
        """
        self.set_io_values(io1=value)

    def set_io2_value(self, value: Optional[NumpySupportedValue]) -> None:
        """
        Sets the values of ``IO2``.

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value: The value to be placed in ``IO2``
        """
        self.set_io_values(io2=value)

    @overload
    def get_io_values(
        self, *, io1_type: None = ..., io2_type: None = ...
    ) -> Tuple[
        GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData,
        GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData,
    ]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: Type[bool], io2_type: None = ...
    ) -> Tuple[bool, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: Type[int], io2_type: None = ...
    ) -> Tuple[int, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: Union[Type[float], Type[fixed]], io2_type: None = ...
    ) -> Tuple[float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: None = ..., io2_type: Type[bool]
    ) -> Tuple[GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, bool]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Type[bool], io2_type: Type[bool]) -> Tuple[bool, bool]:
        pass

    @overload
    def get_io_values(  # type: ignore[overload-overlap]
        self, *, io1_type: Type[int], io2_type: Type[bool]
    ) -> Tuple[int, bool]:
        # This overlap is because int and bool are not distinct types in runtime
        pass

    @overload
    def get_io_values(self, *, io1_type: Union[Type[float], Type[fixed]], io2_type: Type[bool]) -> Tuple[float, bool]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: None = ..., io2_type: Type[int]
    ) -> Tuple[GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, int]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Type[bool], io2_type: Type[int]) -> Tuple[bool, int]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Type[int], io2_type: Type[int]) -> Tuple[int, int]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Union[Type[float], Type[fixed]], io2_type: Type[int]) -> Tuple[float, int]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: None = ..., io2_type: Union[Type[float], Type[fixed]]
    ) -> Tuple[GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData, float]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Type[bool], io2_type: Union[Type[float], Type[fixed]]) -> Tuple[bool, float]:
        pass

    @overload
    def get_io_values(self, *, io1_type: Type[int], io2_type: Union[Type[float], Type[fixed]]) -> Tuple[int, float]:
        pass

    @overload
    def get_io_values(
        self, *, io1_type: Union[Type[float], Type[fixed]], io2_type: Union[Type[float], Type[fixed]]
    ) -> Tuple[float, float]:
        pass

    def get_io_values(
        self, *, io1_type: Optional[IoValueTypes] = None, io2_type: Optional[IoValueTypes] = None
    ) -> Tuple[
        Union[bool, int, float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData],
        Union[bool, int, float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData],
    ]:
        """
        Gets the data stored in ``IO1`` & ``IO2``

        Data will be presented as the type given. If no type was given, it'll have three fields: `int_value`,
        `double_value`, & `boolean_value`
        Args:
            io1_type: The type of ``IO1``
            io2_type:  The type of ``IO1``
        Returns:
             A tuple of (``IO1``, ``IO2``)
        """
        v1, v2 = self._fetch_io_values()
        return _extract_io_value_type(v1, io1_type), _extract_io_value_type(v2, io2_type)

    @overload
    def get_io1_value(
        self,
        as_type: None = ...,
    ) -> GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData:
        pass

    @overload
    def get_io1_value(self, as_type: Type[bool]) -> bool:
        pass

    @overload
    def get_io1_value(self, as_type: Type[int]) -> int:
        pass

    @overload
    def get_io1_value(self, as_type: Union[Type[float], Type[fixed]]) -> float:
        pass

    def get_io1_value(
        self, as_type: Optional[IoValueTypes] = None
    ) -> Union[bool, int, float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
        """
        Gets the data stored in ``IO1``

        Data will be presented as the type given. If no type was given, it'll have three fields: `int_value`,
        `double_value`, & `boolean_value`

        Args:
            as_type: The type of ``IO1``
        Returns:
             ``IO1``
        """
        return self.get_io_values(io1_type=as_type)[0]

    @overload
    def get_io2_value(
        self,
        as_type: None = ...,
    ) -> GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData:
        pass

    @overload
    def get_io2_value(self, as_type: Type[bool]) -> bool:
        pass

    @overload
    def get_io2_value(self, as_type: Type[int]) -> int:
        pass

    @overload
    def get_io2_value(self, as_type: Union[Type[float], Type[fixed]]) -> float:
        pass

    def get_io2_value(
        self, as_type: Optional[IoValueTypes] = None
    ) -> Union[bool, int, float, GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData]:
        """
        Gets the data stored in ``IO12`

        Data will be presented as the type given. If no type was given, it'll have three fields: `int_value`,
        `double_value`, & `boolean_value`

        Args:
            as_type: The type of ``IO2``
        Returns:
             ``IO2``
        """
        return self.get_io_values(io2_type=as_type)[1]

    def _fetch_io_values(
        self,
    ) -> Tuple[
        GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData,
        GetIoValuesResponseGetIoValuesResponseSuccessIoValuesData,
    ]:
        request = GetIoValuesRequest(job_id=self._id)
        response = self._run(self._stub.get_io_values(request, timeout=self._timeout))
        return response.io1, response.io2

    @property
    def result_handles(self) -> StreamingResultFetcher:
        return StreamingResultFetcher(
            self._id,
            JobResultApi(self.connection_details, self._id),
            self._store,
            self._caps,
        )

    def get_errors(self) -> List[ExecutionError]:
        """
        Returns:

             A list of all errors in the execution report
        """
        request = GetJobErrorsRequest(job_id=self._id)
        response = self._run(self._stub.get_errors(request, timeout=self._timeout))
        return [ExecutionError.create_from_grpc_message(error) for error in response.errors]

    def execution_report(self) -> ExecutionReport:
        """Get runtime errors report for this job. See [Runtime errors](../Guides/error.md#runtime-errors).

        Returns:
            An object holding the errors that this job generated.
        """
        errors = self.get_errors()
        return ExecutionReport(self._id, errors)

    def set_element_correction(self, element: str, correction: Tuple[float, float, float, float]) -> None:
        r"""Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with an element.

        Values will be rounded to an accuracy of $2^{-16}$.
        Valid values for the correction values are between $-2$ and $(2 - 2^{-16})$.

        Warning - the correction matrix can increase the output voltage which might result in an
        overflow.

        Args:
            element (str): The name of the element to update the correction for
            correction (tuple):
                Tuple is of the form (v00, v01, v10, v11) where
                the matrix is
                $\begin{pmatrix} v_{00} & v_{01} \\ v_{10} & v_{11}\end{pmatrix}$
        """
        element_input = self.elements[element].input
        if not isinstance(element_input, MixInputsApi):
            raise ValueError(f"Element {element} is not an IQ mixer")
        element_input.set_correction(correction)

    def get_element_correction(self, element: str) -> Tuple[float, float, float, float]:
        """Gets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with an element.

        Args:
            element (str): The name of the element to update the correction for

        Returns:
            The current correction matrix
        """
        element_input = self.elements[element].input
        if not isinstance(element_input, MixInputsApi):
            raise ValueError(f"Element {element} is not an IQ mixer")
        return element_input.get_correction()

    def set_intermediate_frequency(self, element: str, freq: float) -> None:
        """Sets the intermediate frequency of the element

        Args:
            element (str): The name of the element whose intermediate frequency will be updated
            freq (float): The intermediate frequency to set to the given element
        """
        element_inst = self.elements[element]
        element_inst.set_intermediate_frequency(freq)

    def get_intermediate_frequency(self, element: str) -> float:
        """Gets the intermediate frequency of the element

        Args:
            element (str): The name of the element whose intermediate frequency will be updated

        Returns:
            float: The intermediate frequency of the given element
        """
        element_inst = self.elements[element]
        return element_inst.get_intermediate_frequency()

    def get_output_dc_offset_by_element(
        self, element: str, iq_input: Optional[Literal["I", "Q", "single"]] = None
    ) -> float:
        """Get the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element: The name of the element to get the correction for
            iq_input: The port name as appears in the element config.
                Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'`
                    for an element with mixer inputs

        Returns:
            The offset, in volts
        """
        input_instance = self.elements[element].input
        if isinstance(input_instance, SingleInputApi):
            return input_instance.get_dc_offset()
        elif isinstance(input_instance, MixInputsApi):
            i, q = input_instance.get_dc_offsets()
            if iq_input == "I":
                return i
            elif iq_input == "Q":
                return q
            else:
                raise FunctionInputError(f"Port must be I or Q, got {iq_input}.")
        else:
            raise ValueError(f"Element {element} of type {type(input_instance)} does not have a 'port' property.")

    @overload
    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Union[Tuple[Literal["I", "Q"], Literal["I", "Q"]], List[Literal["I", "Q"]]],
        offset: Union[Tuple[float, float], List[float]],
    ) -> None:
        pass

    @overload
    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Literal["single", "I", "Q"],
        offset: float,
    ) -> None:
        pass

    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Union[Literal["single", "I", "Q"], Tuple[Literal["I", "Q"], Literal["I", "Q"]], List[Literal["I", "Q"]]],
        offset: Union[float, Tuple[float, float], List[float]],
    ) -> None:
        """Set the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element (str): the name of the element to update the correction for
            input (Union[str, Tuple[str,str], List[str]]): the input name as appears in the element config. Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'` or a tuple ('I', 'Q')
                    for an element with mixer inputs
            offset (Union[float, Tuple[float,float], List[float]]): The dc value to set to, in volts. Ranges
                from -0.5 to 0.5 - 2^-16 in steps of 2^-16.

        Examples:
            ```python
            job.set_output_dc_offset_by_element('flux', 'single', 0.1)
            job.set_output_dc_offset_by_element('qubit', 'I', -0.01)
            job.set_output_dc_offset_by_element('qubit', ('I', 'Q'), (-0.01, 0.05))
            ```

        Note:

            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        input_instance = self.elements[element].input
        if isinstance(input_instance, MixInputsApi):
            if isinstance(input, (list, tuple)):
                if not set(input) <= {"I", "Q"}:
                    raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
                if not (isinstance(offset, (list, tuple)) and len(input) == len(offset)):
                    raise FunctionInputError(
                        f"input should be two iterables of the same size," f"got input = {input} and offset = {offset}"
                    )
                channel_to_offset = dict(zip(input, offset))
            elif isinstance(input, str):
                if input not in {"I", "Q"}:
                    raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
                if not isinstance(offset, (int, float)):
                    raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
                channel_to_offset = {cast(Literal["I", "Q"], input): offset}
            else:
                raise ValueError(f"Invalid input - {input}")
            input_instance.set_dc_offsets(i_offset=channel_to_offset.get("I"), q_offset=channel_to_offset.get("Q"))
        elif isinstance(input_instance, SingleInputApi):
            if input != "single":
                raise ValueError(f"Invalid input - {input} while the element has a single input")
            if not isinstance(offset, (int, float)):
                raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
            input_instance.set_dc_offset(offset)
        else:
            raise ValueError(
                f"Element {element} with input of type {input_instance.__class__.__name__} "
                f"does not support dc offset setting."
            )

    def set_input_dc_offset_by_element(self, element: str, output: str, offset: float) -> None:
        """Set the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element (str): the name of the element to update the
                correction for
            output (str): the output key name as appears in the element
                config under 'outputs'.
            offset (float): the dc value to set to, in volts. Ranges from -0.5 to 0.5 - 2^-16 in steps of
                2^-16.

        Note:
            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        element_instance = self.elements[element]
        element_instance.outputs[output].set_dc_offset(offset)

    def get_input_dc_offset_by_element(self, element: str, output: str) -> float:
        """Get the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element: the name of the element to get the correction for
            output: the output key name as appears in the element config
                under 'outputs'.

        Returns:
            The offset, in volts
        """
        element_instance = self.elements[element]
        return element_instance.outputs[output].get_dc_offset()

    def get_output_digital_delay(self, element: str, digital_input: str) -> int:
        """Gets the delay of the digital input of the element

        Args:
            element: The name of the element to get the delay for
            digital_input: The digital input name as appears in the
                element's config

        Returns:
            The delay
        """
        element_instance = self.elements[element]
        return element_instance.digital_inputs[digital_input].get_delay()

    def set_output_digital_delay(self, element: str, digital_input: str, delay: int) -> None:
        """Sets the delay of the digital input of the element

        Args:
            element (str): The name of the element to update delay for
            digital_input (str): The digital input name as appears in
                the element's config
            delay (int): The delay value to set to, in ns.
        """
        element_instance = self.elements[element]
        element_instance.digital_inputs[digital_input].set_delay(delay)

    def get_output_digital_buffer(self, element: str, digital_input: str) -> int:
        """Gets the buffer for digital input of the element

        Args:
            element (str): The name of the element to get the buffer for
            digital_input (str): The digital input name as appears in
                the element's config

        Returns:
            The buffer
        """
        element_instance = self.elements[element]
        return element_instance.digital_inputs[digital_input].get_buffer()

    def set_output_digital_buffer(self, element: str, digital_input: str, buffer: int) -> None:
        """Sets the buffer for digital input of the element

        Args:
            element (str): The name of the element to update buffer for
            digital_input (str): the digital input name as appears in
                the element's config
            buffer (int): The buffer value to set to, in ns.
        """
        element_instance = self.elements[element]
        element_instance.digital_inputs[digital_input].set_buffer(buffer)

    def update_oscillator_frequency(
        self,
        element: str,
        frequency_hz: float,
        update_component: Literal["upconverter", "downconverter", "both"] = "both",
    ) -> None:
        """Set the upconverter frequency or downconverter frequency of the microwave input of the element

        Args:
            element (str): The name of the element to update the correction for
            frequency_hz (float): The frequency to set to the given element
            update_component (str): The component to update the frequency for: "upconverter", "downconverter", or "both"
        """
        if update_component == "upconverter":
            input_instance = self.elements[element].input
            if not isinstance(input_instance, MwInputApi):
                raise ValueError(f"Element {element} does not have a microwave input")
            input_instance.set_oscillator_frequency(frequency_hz, set_also_output=False)
        elif update_component == "downconverter":
            output_instance = self.elements[element].microwave_output
            output_instance.set_oscillator_frequency(frequency_hz)
        elif update_component == "both":
            input_instance = self.elements[element].input
            if isinstance(input_instance, MwInputApi):
                input_instance.set_oscillator_frequency(
                    frequency_hz, set_also_output=self.elements[element].has_mw_output
                )
            else:
                output_instance = self.elements[element].microwave_output
                output_instance.set_oscillator_frequency(frequency_hz)
        else:
            raise ValueError(f"Invalid update_component: {update_component}")


class JobApiWithDeprecations(JobApi):
    @property
    def status(self) -> str:
        """Returns the status of the job, one of the following strings:
        "unknown", "pending", "running", "completed", "canceled", "loading", "error"
        """
        new_status = self.get_status()
        old_api_status = {
            "In queue": "pending",
            "Running": "running",
            "Processing": "running",
            "Done": "completed",
            "Canceled": "canceled",
            "Error": "error",
        }[new_status]
        return old_api_status

    @property
    def user_added(self) -> None:
        return None

    @property
    def time_added(self) -> None:
        return None

    def insert_input_stream(
        self,
        name: str,
        data: List[Value],
    ) -> None:
        """Deprecated - Please use `job.push_to_input_stream`."""
        warnings.warn(
            deprecation_message(
                method="job.insert_input_stream",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method was renamed to `job.push_to_input_stream()`.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self.push_to_input_stream(name, data)

    def push_to_input_stream(self, name: str, data: List[Value]) -> None:
        """Push data to the input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.

        Multiple data entries can be inserted before the data is read by the program.

        See [Input streams](../Guides/features.md#input-streams) for more information.

        Args:
            name: The input stream name the data is to be inserted to.
            data: The data to be inserted. The data's size must match
                the size of the input stream.
        """
        if not isinstance(data, list):
            data = [data]
        super().push_to_input_stream(name, data)

    def halt(self) -> bool:
        """Halts the job on the opx"""
        warnings.warn(
            deprecation_message(
                method="job.halt",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method was renamed to `job.cancel`.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self.cancel()
        return True

    def wait_for_execution(self, timeout: float = float("infinity")) -> "JobApi":
        """Deprecated - This method is going to be removed, please use `job.wait_until("Running")`.

        Waits until the job has passed the "Running" state.
        If the timeout is reached, the function will raise an error.
        Args:
            timeout: The timeout time, in seconds
        Returns:
             The running job
        """
        warnings.warn(
            deprecation_message(
                method="job.wait_for_execution",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details='This method is going to be removed, please use `job.wait_until("Running")`.',
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self.wait_until({"Running"}, timeout)
        return self



import logging
from typing import Union, Generic

import numpy
from dependency_injector.wiring import Provide, inject

from qm.api.frontend_api import FrontendApi
from qm.elements.element_outputs import ElementOutput
from qm.api.models.capabilities import ServerCapabilities
from qm.elements.up_converted_input import UpconvertedInputNewApi
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.elements.element_inputs import ElementInput, ElementInputGRPCType
from qm.grpc.qua_config import (
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigSingleInput,
    QuaConfigMultipleInputs,
    QuaConfigSingleInputCollection,
)

logger = logging.getLogger(__name__)


class Element(Generic[ElementInputGRPCType]):
    def __init__(
        self,
        name: str,
        config: QuaConfigElementDec,
        api: FrontendApi,
        machine_id: str,
        element_input: ElementInput[ElementInputGRPCType],
        element_output: ElementOutput,
    ):
        self._config = config
        self._name = name
        self._frontend = api
        self._id = machine_id
        self.input: ElementInput[ElementInputGRPCType] = element_input
        self.output = element_output

    @property
    def name(self) -> str:
        return self._name

    @inject
    def set_intermediate_frequency(
        self, freq: float, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
    ) -> None:
        if not isinstance(freq, (numpy.floating, float)):
            raise TypeError("freq must be a float")

        freq = float(freq)
        logger.debug(f"Setting element '{self._name}' intermediate frequency to '{freq}'.")
        self._frontend.set_intermediate_frequency(self._id, self._name, freq)
        self._config.intermediate_frequency_double = float(freq) if capabilities.supports_double_frequency else 0.0
        self._config.intermediate_frequency = int(freq)

    @property
    @inject
    def intermediate_frequency(
        self, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
    ) -> float:
        sign: int = (-1) ** self._config.intermediate_frequency_negative
        if capabilities.supports_double_frequency:
            freq = self._config.intermediate_frequency_double
        else:
            freq = float(self._config.intermediate_frequency or 0)
        return sign * freq

    def get_output_digital_delay(self, digital_input_name: str) -> int:
        try:
            return self._config.digital_inputs[digital_input_name].delay
        except KeyError:
            raise Exception(f"Digital input for {digital_input_name} was not found.")

    def get_output_digital_buffer(self, digital_input_name: str) -> int:
        try:
            return self._config.digital_inputs[digital_input_name].buffer
        except KeyError:
            raise Exception(f"Digital buffer for {digital_input_name} was not found.")

    def set_output_digital_delay(self, digital_input: str, delay: int) -> None:
        if not isinstance(digital_input, str):
            raise Exception("port must be a string")
        if not isinstance(delay, int):
            raise Exception("delay must be an int")
        logger.debug(f"Setting delay of digital port '{digital_input}' on element '{self._name}' to '{delay}'")
        self._frontend.set_output_digital_delay(self._id, self._name, digital_input, delay)
        self._config.digital_inputs[digital_input].delay = delay

    def set_output_digital_buffer(self, digital_input: str, buffer: int) -> None:
        if not isinstance(digital_input, str):
            raise Exception("port must be a string")
        if not isinstance(buffer, int):
            raise Exception("buffer must be an int.")
        logger.debug(f"Setting buffer of digital port '{digital_input}' on element '{self._name}' to '{buffer}'")
        self._frontend.set_output_digital_buffer(self._id, self._name, digital_input, buffer)
        self._config.digital_inputs[digital_input].buffer = buffer

    def set_input_dc_offset(self, output: str, offset: float) -> None:
        logger.debug(f"Setting DC offset of output '{output}' on element '{self._name}' to '{offset}'.")
        if not isinstance(output, str):
            raise TypeError("output must be a string")
        if output not in self._config.outputs:
            raise ValueError(f"output {output} was not found in the element's outputs")
        if not isinstance(offset, (numpy.floating, float)):
            raise TypeError("offset must be a float")
        offset = float(offset)
        self._frontend.set_input_dc_offset(self._id, self._name, output, offset)

    @property
    def time_of_flight(self) -> int:
        return self._config.time_of_flight or 0

    @property
    def smearing(self) -> int:
        return self._config.smearing or 0


class NewApiUpconvertedElement(Element[QuaConfigMixInputs]):
    def __init__(
        self,
        name: str,
        config: QuaConfigElementDec,
        element_input: UpconvertedInputNewApi,
        element_output: ElementOutput,
    ):
        super().__init__(
            name,
            config,
            api=None,  # type: ignore[arg-type]
            machine_id=None,  # type: ignore[arg-type]
            element_input=element_input,
            element_output=element_output,
        )

    @property
    def name(self) -> str:
        return self._name

    @inject
    def set_intermediate_frequency(
        self, freq: float, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
    ) -> None:
        raise NotImplementedError

    def set_output_digital_delay(self, digital_input: str, delay: int) -> None:
        raise NotImplementedError

    def set_output_digital_buffer(self, digital_input: str, buffer: int) -> None:
        raise NotImplementedError

    def set_input_dc_offset(self, output: str, offset: float) -> None:
        raise NotImplementedError


AllElements = Union[
    Element[QuaConfigSingleInput],
    Element[QuaConfigMixInputs],
    Element[QuaConfigSingleInputCollection],
    Element[QuaConfigMultipleInputs],
    Element[None],
]



from collections import defaultdict
from typing import Dict, List, Union, Optional, Sequence

import betterproto

from qm.api.v2.job_api import JobApi
from qm.persistence import BaseStore
from qm.utils.async_utils import run_async
from qm.exceptions import QMSimulationError
from qm.waveform_report import WaveformReport
from qm.utils.config_utils import get_fem_config
from qm.grpc.frontend import SimulatedResponsePart
from qm.jobs.simulated_job import extract_struct_value
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.job_api.job_api import JobApiWithDeprecations
from qm.results.simulator_samples import SimulatorSamples, SimulatorControllerSamples
from qm.grpc.qua_config import (
    QuaConfigControllerDec,
    QuaConfigOctoDacFemDec,
    QuaConfigMicrowaveFemDec,
    QuaConfigAdcPortReference,
)
from qm.grpc.v2 import (
    PullSamplesRequest,
    PullSamplesResponsePullSamplesResponseSuccess,
    PullSamplesResponsePullSamplesResponseSuccessLf,
    PullSamplesResponsePullSamplesResponseSuccessMw,
    PullSamplesResponsePullSamplesResponseSuccessMode,
)


class SimulatedJobApi(JobApi):
    def __init__(
        self,
        connection_details: ConnectionDetails,
        job_id: str,
        store: BaseStore,
        simulated_response: SimulatedResponsePart,
        capabilities: ServerCapabilities,
    ) -> None:
        super().__init__(connection_details, job_id, store, capabilities)
        self._simulated_response = simulated_response

        self._waveform_report = WaveformReport.from_dict(
            extract_struct_value(simulated_response.waveform_report), self.id
        )

    def get_simulated_waveform_report(self) -> Optional[WaveformReport]:
        return self._waveform_report

    async def _pull_simulator_samples(
        self, include_analog: bool, include_digital: bool
    ) -> Dict[str, List[PullSamplesResponsePullSamplesResponseSuccess]]:
        request = PullSamplesRequest(self._id, include_analog, include_digital)
        bare_results = defaultdict(list)
        async for result in self._stub.pull_samples(request, timeout=self._timeout):
            _, response = betterproto.which_one_of(result, "response_oneof")
            if isinstance(response, PullSamplesResponsePullSamplesResponseSuccess):
                bare_results[response.controller].append(response)
            else:
                raise QMSimulationError("Error while pulling samples")
        return dict(bare_results)

    def get_simulated_samples(self, include_analog: bool = True, include_digital: bool = True) -> SimulatorSamples:
        config = self._get_pb_config()

        results_by_controller = run_async(self._pull_simulator_samples(include_analog, include_digital))
        controller_to_samples = {}
        for controller, responses in results_by_controller.items():
            analog: Dict[str, Sequence[Union[float, complex]]] = {}
            digital = {}
            analog_sampling_rate = {}
            for response in responses:
                fem_config = get_fem_config(
                    config,
                    QuaConfigAdcPortReference(controller=controller, fem=response.fem_id, number=response.port_id),
                )
                key = f"{response.fem_id}-{response.port_id}"
                if response.mode == PullSamplesResponsePullSamplesResponseSuccessMode.ANALOG:
                    if isinstance(fem_config, QuaConfigControllerDec):
                        analog_sampling_rate[key] = 1e9
                    elif isinstance(fem_config, QuaConfigOctoDacFemDec):
                        analog_sampling_rate[key] = 2e9
                    elif isinstance(fem_config, QuaConfigMicrowaveFemDec):
                        analog_sampling_rate[key + "-1"] = fem_config.analog_outputs[response.port_id].sampling_rate
                        analog_sampling_rate[key + "-2"] = fem_config.analog_outputs[response.port_id].sampling_rate
                    else:
                        raise QMSimulationError(f"Unknown FEM type: {fem_config}")

                    samples = response.double_data
                    _, data = betterproto.which_one_of(samples, "output")
                    if isinstance(data, PullSamplesResponsePullSamplesResponseSuccessMw):
                        analog[f"{key}-{data.duc_id}"] = [x + 1j * y for x, y in zip(data.i, data.q)]
                    elif isinstance(data, PullSamplesResponsePullSamplesResponseSuccessLf):
                        analog[key] = data.items
                else:
                    digital[key] = response.boolean_data.data[0].item
            controller_to_samples[controller] = SimulatorControllerSamples(
                analog, digital, analog_sampling_rate=analog_sampling_rate
            )
        return SimulatorSamples(controller_to_samples)

    def plot_waveform_report_with_simulated_samples(self) -> None:
        self._waveform_report.create_plot(self.get_simulated_samples())

    def plot_waveform_report_without_samples(self) -> None:
        self._waveform_report.create_plot()


class SimulatedJobApiWithDeprecations(JobApiWithDeprecations, SimulatedJobApi):
    pass



import ssl
import dataclasses
from typing import Dict, Tuple, Optional

from qm.api.models.info import QuaMachineInfo
from qm.api.models.debug_data import DebugData
from qm.api.models.capabilities import ServerCapabilities

MAX_MESSAGE_SIZE = 1024 * 1024 * 100  # 100 mb in bytes
BASE_TIMEOUT = 60


@dataclasses.dataclass
class ConnectionDetails:
    host: str
    port: int
    user_token: Optional[str] = dataclasses.field(default=None)
    ssl_context: Optional[ssl.SSLContext] = dataclasses.field(default=None)
    max_message_size: int = dataclasses.field(default=MAX_MESSAGE_SIZE)
    headers: Dict[str, str] = dataclasses.field(default_factory=dict)
    timeout: float = dataclasses.field(default=BASE_TIMEOUT)
    debug_data: Optional[DebugData] = dataclasses.field(default=None)

    def __hash__(self) -> int:
        return hash(
            (
                self.host,
                self.port,
                self.user_token,
                self.ssl_context,
                self.max_message_size,
                tuple(sorted(self.headers.items())),
                self.timeout,
                self.debug_data,
            )
        )


@dataclasses.dataclass
class ResponseConnectionDetails:
    host: str
    port: int
    octaves: Dict[str, Tuple[str, int]]


@dataclasses.dataclass
class ServerDetails:
    port: int
    host: str
    server_version: str
    connection_details: ConnectionDetails
    octaves: Dict[str, ConnectionDetails]

    # does it implement the QUA service
    qua_implementation: Optional[QuaMachineInfo]
    capabilities: ServerCapabilities = dataclasses.field(default_factory=ServerCapabilities.build)

    def __post_init__(self) -> None:
        self.capabilities = ServerCapabilities.build(qua_implementation=self.qua_implementation)



import logging
import warnings
from dataclasses import field, dataclass
from typing import Dict, List, Type, Union, Literal, Mapping, Iterable, Optional

from qm.grpc.qua import QuaProgram
from qm.utils import LOG_LEVEL_MAP
from qm.octave import QmOctaveConfig
from qm.persistence import BaseStore
from qm.grpc.qua_config import QuaConfig
from qm.api.v2.base_api_v2 import BaseApiV2
from qm.octave.octave_manager import OctaveManager
from qm.api.v2.qm_api_old import QmApiWithDeprecations
from qm.type_hinting.config_types import DictQuaConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.qm_api import QmApi, handle_simulation_error
from qm.exceptions import OpenQmException, QopResponseError
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi
from qm.grpc.frontend import InterOpxConnection, ExecutionRequestSimulate
from qm.api.v2.job_api.job_api import JobApi, JobData, JobStatus, JobApiWithDeprecations, transfer_statuses_to_enum
from qm.grpc.v2 import (
    GetJobsRequest,
    QmmServiceStub,
    JobsQueryParams,
    GetVersionRequest,
    HealthCheckRequest,
    GetControllersRequest,
    ClearAllJobResultsRequest,
    OpenQuantumMachineRequest,
    QmmServiceSimulateRequest,
    CloseAllQuantumMachinesRequest,
    ListOpenQuantumMachinesRequest,
    OpenQuantumMachineRequestCloseMode,
    QmmServiceResetDataProcessingRequest,
)

logger = logging.getLogger(__name__)


@dataclass
class VersionResponse:
    gateway: str
    controllers: Dict[str, str]


ControllerTypes = Literal["OPX", "OPX1000"]
FemTypes = Literal["LF", "MW"]
FEM_TYPES_MAPPING: Dict[int, FemTypes] = {1: "LF", 2: "MW"}


@dataclass
class ControllerBase:
    name: str

    @property
    def temperature(self) -> Optional[float]:
        return None

    @property
    def controller_type(self) -> ControllerTypes:
        raise NotImplementedError


@dataclass
class Controller(ControllerBase):
    _temperature: Optional[float] = field(repr=False)

    @property
    def temperature(self) -> Optional[float]:
        if self._temperature is None:
            logger.warning(
                "Temperature is not available for this controller, "
                "Either because it uses an old gateway or its state is not known."
            )
        return self._temperature

    @property
    def controller_type(self) -> ControllerTypes:
        return "OPX"


@dataclass
class ControllerOPX1000(ControllerBase):
    hostname: str
    fems: Dict[int, FemTypes]

    @property
    def controller_type(self) -> ControllerTypes:
        return "OPX1000"

    @property
    def temperature(self) -> Optional[float]:
        logger.warning("Temperature is not yet available for OPX1000 controllers.")
        return None


class QmmApi(BaseApiV2[QmmServiceStub]):
    QM_CLASS = QmApi
    JOB_CLASS = JobApi

    def __init__(
        self,
        connection_details: ConnectionDetails,
        store: BaseStore,
        capabilities: ServerCapabilities,
        octave_config: Optional[QmOctaveConfig],
        octave_manager: OctaveManager,
    ) -> None:
        super().__init__(connection_details)
        self._store = store
        self._caps = capabilities
        self._octave_config = octave_config
        self._octave_manager = octave_manager

    @property
    def _stub_class(self) -> Type[QmmServiceStub]:
        return QmmServiceStub

    def get_qm(self, qm_id: str, _pb_config: Optional[QuaConfig] = None) -> QmApi:
        # todo - remove _pb_config when octave config is in the GW
        return self.QM_CLASS(
            connection_details=self.connection_details,
            qm_id=qm_id,
            store=self._store,
            capabilities=self._caps,
            octave_config=self._octave_config,
            octave_manager=self._octave_manager,
            pb_config=_pb_config,
        )

    def get_job(self, job_id: str) -> JobApi:
        return self.JOB_CLASS(self.connection_details, job_id, store=self._store, capabilities=self._caps)

    def get_jobs(
        self,
        qm_ids: Iterable[str] = tuple(),
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Union[JobStatus, Iterable[JobStatus]] = tuple(),
    ) -> List[JobData]:
        query_params = JobsQueryParams(
            quantum_machine_ids=list(qm_ids),
            job_ids=list(job_ids),
            user_ids=list(user_ids),
            description=description,
            status=transfer_statuses_to_enum(status),
        )
        request = GetJobsRequest(query=query_params)
        response = self._run(self._stub.get_jobs(request, timeout=self._timeout))
        return [JobData.from_grpc(j) for j in response.jobs]

    def open_qm(self, config: QuaConfig, close_other_machines: Optional[bool] = None) -> QmApi:
        if close_other_machines is None:
            warnings.warn(
                "close_other_machines is not set, future default will be False, now setting to True", DeprecationWarning
            )
            close_other_machines = True
        request = OpenQuantumMachineRequest(
            config=config,
            close_mode=OpenQuantumMachineRequestCloseMode.CLOSE_MODE_IF_NEEDED
            if close_other_machines
            else OpenQuantumMachineRequestCloseMode.CLOSE_MODE_UNSPECIFIED,
        )

        try:
            response = self._run(self._stub.open_quantum_machine(request, timeout=self._timeout))

        except QopResponseError as e:
            error = e.error
            error_messages = []
            for sub_error in error.config_validation_errors:
                error_messages.append(
                    f'CONFIG ERROR in key "{sub_error.path}" [{sub_error.group}] : {sub_error.message}'
                )

            for physical_error in error.physical_validation_errors:
                error_messages.append(
                    f'PHYSICAL CONFIG ERROR in key "{physical_error.path}" [{physical_error.group}] : {physical_error.message}'
                )

            for msg in error_messages:
                logger.error(msg)

            error_details = [(item.group, item.path, item.message) for item in error.config_validation_errors] + [
                (item.group, item.path, item.message) for item in error.physical_validation_errors
            ]
            formatted_errors = "\n".join(error_messages)
            raise OpenQmException(
                f"Can not open QM, see the following errors:\n{formatted_errors}", errors=error_details
            )

        for warning in response.open_qm_warnings:
            logger.warning(f"Open QM ended with warning {warning.code}: {warning.message}")

        return self.get_qm(response.quantum_machine_id, _pb_config=config)

    def perform_healthcheck(self) -> None:
        logger.info("Performing health check")
        response = self._run(self._stub.health_check(HealthCheckRequest(), timeout=self._timeout))
        msg = "Cluster healthcheck completed successfully."
        if response.details:
            msg += " Details:"
            for k, v in response.details.items():
                msg += f"\n  {k}: {v}"
        logger.info(msg)

    def get_version(self) -> VersionResponse:
        response = self._run(self._stub.get_version(GetVersionRequest(), timeout=self._timeout))
        return VersionResponse(
            gateway=response.gateway,
            controllers={k: v for k, v in response.controllers.items()},
        )

    def get_controllers(self) -> Mapping[str, ControllerBase]:
        response = self._run(self._stub.get_controllers(GetControllersRequest(), timeout=self._timeout))
        to_return = {}
        correction_offset = 0 if self._caps.opx1000_fems_return_1_based else 1
        for name, value in response.control_devices.items():
            if value.controller_type == 1:
                to_return[name] = ControllerOPX1000(
                    name=name,
                    hostname=value.hostname,
                    fems={
                        int(i) + correction_offset: FEM_TYPES_MAPPING[f.type]
                        for i, f in value.fems.items()
                        if f.type > 0
                    },
                )
            else:
                raise NotImplementedError(f"Controller type {value.controller_type} is not supported.")
        return to_return

    def reset_data_processing(self) -> None:
        request = QmmServiceResetDataProcessingRequest()
        self._run(self._stub.reset_data_processing(request, timeout=self._timeout))

    def validate_qua_config(self, config: DictQuaConfig) -> None:
        pass

    def simulate(
        self,
        config: QuaConfig,
        program: QuaProgram,
        simulate: ExecutionRequestSimulate,
        controller_connections: List[InterOpxConnection],
    ) -> SimulatedJobApi:
        request = QmmServiceSimulateRequest(
            config=config,
            high_level_program=program,
            simulate=simulate,
            controller_connections=controller_connections,
        )
        logger.info("Simulating program.")
        try:
            response = self._run(self._stub.simulate(request, timeout=self._timeout))
        except QopResponseError as e:
            raise handle_simulation_error(e)

        for msg in response.messages:
            lvl = LOG_LEVEL_MAP[msg.level]
            logger.log(lvl, msg.message)

        return self.QM_CLASS.SIMULATED_JOB_CLASS(
            self.connection_details,
            response.job_id,
            store=self._store,
            simulated_response=response.simulated,
            capabilities=self._caps,
        )

    def list_open_qms(self) -> List[str]:
        request = ListOpenQuantumMachinesRequest()
        response = self._run(self._stub.list_open_quantum_machines(request, timeout=self._timeout))
        return response.machine_ids

    def close_all_qms(self) -> None:
        request = CloseAllQuantumMachinesRequest()
        self._run(self._stub.close_all_quantum_machines(request, timeout=self._timeout))

    def clear_all_job_results(self) -> None:
        request = ClearAllJobResultsRequest()
        self._run(self._stub.clear_all_job_results(request, timeout=self._timeout))


class QmmApiWithDeprecations(QmmApi):
    QM_CLASS = QmApiWithDeprecations
    JOB_CLASS = JobApiWithDeprecations



from typing import Type, Tuple, Union, Optional, overload

from qm.grpc.qm_api import Matrix
from qm.api.v2.job_api.generic_apis import ElementGenericApi
from qm.grpc.qua_config import (
    QuaConfigMixInputs,
    QuaConfigSingleInput,
    QuaConfigMultipleInputs,
    QuaConfigSingleInputCollection,
    QuaConfigMicrowaveInputPortReference,
)
from qm.grpc.v2 import (
    MixInputsDcOffset,
    SingleInputDcOffset,
    GetOutputDcOffsetRequest,
    SetOutputDcOffsetRequest,
    GetMatrixCorrectionRequest,
    SetMatrixCorrectionRequest,
    SetOscillatorFrequencyRequest,
    SetOscillatorFrequencyRequestUpdateComponentSelection,
)

InputConfigType = Union[
    QuaConfigSingleInput,
    QuaConfigMixInputs,
    QuaConfigSingleInputCollection,
    QuaConfigMultipleInputs,
    QuaConfigMicrowaveInputPortReference,
    None,
]


class UnknownElementType(ValueError):
    pass


class ElementInputApi(ElementGenericApi):
    pass


class NoInputApi(ElementInputApi):
    pass


class MultipleInputsApi(ElementInputApi):
    pass


class SingleInputCollectionApi(ElementInputApi):
    pass


class MixInputsApi(ElementInputApi):
    def set_correction(self, correction: Tuple[float, float, float, float]) -> None:
        matrix = Matrix(v00=correction[0], v01=correction[1], v10=correction[2], v11=correction[3])
        request = SetMatrixCorrectionRequest(job_id=self._id, qe=self._element_id, correction=matrix)
        self._run(self._stub.set_matrix_correction(request, timeout=self._timeout))

    def get_correction(self) -> Tuple[float, float, float, float]:
        request = GetMatrixCorrectionRequest(job_id=self._id, qe=self._element_id)
        correction = self._run(self._stub.get_matrix_correction(request, timeout=self._timeout)).correction
        return correction.v00, correction.v01, correction.v10, correction.v11

    def set_dc_offsets(self, i_offset: Optional[float] = None, q_offset: Optional[float] = None) -> None:
        request = SetOutputDcOffsetRequest(
            job_id=self._id, qe=self._element_id, mix_inputs=MixInputsDcOffset(i=i_offset, q=q_offset)
        )
        self._run(self._stub.set_output_dc_offset(request, timeout=self._timeout))

    def get_dc_offsets(self) -> Tuple[float, float]:
        request = GetOutputDcOffsetRequest(job_id=self._id, qe=self._element_id)
        response = self._run(self._stub.get_output_dc_offset(request, timeout=self._timeout))
        i, q = response.mix_inputs.i, response.mix_inputs.q
        if i is None or q is None:
            raise ValueError("Mixed DC offsets are not set")
        return i, q


class SingleInputApi(ElementInputApi):
    def set_dc_offset(self, offset: float) -> None:
        request = SetOutputDcOffsetRequest(
            job_id=self._id, qe=self._element_id, single_input=SingleInputDcOffset(offset=offset)
        )
        self._run(self._stub.set_output_dc_offset(request, timeout=self._timeout))

    def get_dc_offset(self) -> float:
        request = GetOutputDcOffsetRequest(job_id=self._id, qe=self._element_id)
        response = self._run(self._stub.get_output_dc_offset(request, timeout=self._timeout))
        return response.single_input.offset


class MwInputApi(ElementInputApi):
    def set_oscillator_frequency(self, frequency_hz: float, set_also_output: bool = True) -> None:
        request = SetOscillatorFrequencyRequest(
            job_id=self._id,
            qe=self._element_id,
            new_frequency_hz=frequency_hz,
            update_component=SetOscillatorFrequencyRequestUpdateComponentSelection.both
            if set_also_output
            else SetOscillatorFrequencyRequestUpdateComponentSelection.upconverter,
        )
        self._run(self._stub.set_oscillator_frequency(request, timeout=self._timeout))


@overload
def create_element_input_class(input_config: QuaConfigSingleInput) -> Type[SingleInputApi]:
    pass


@overload
def create_element_input_class(input_config: QuaConfigMixInputs) -> Type[MixInputsApi]:
    pass


@overload
def create_element_input_class(input_config: QuaConfigSingleInputCollection) -> Type[SingleInputCollectionApi]:
    pass


@overload
def create_element_input_class(input_config: QuaConfigMultipleInputs) -> Type[MultipleInputsApi]:
    pass


@overload
def create_element_input_class(input_config: QuaConfigMicrowaveInputPortReference) -> Type[MwInputApi]:
    pass


@overload
def create_element_input_class(input_config: None) -> Type[NoInputApi]:
    pass


def create_element_input_class(input_config: InputConfigType) -> Type[ElementInputApi]:
    if isinstance(input_config, QuaConfigSingleInput):
        return SingleInputApi
    if isinstance(input_config, QuaConfigMixInputs):
        return MixInputsApi
    if isinstance(input_config, QuaConfigSingleInputCollection):
        return SingleInputCollectionApi
    if isinstance(input_config, QuaConfigMultipleInputs):
        return MultipleInputsApi
    if isinstance(input_config, QuaConfigMicrowaveInputPortReference):
        return MwInputApi
    if input_config is None:
        return NoInputApi
    raise UnknownElementType(f"Unknown input type {type(input_config)}")



import base64
import logging
import os.path
import datetime
import dataclasses
from copy import deepcopy
from dataclasses import dataclass
from collections import defaultdict
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Set,
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Mapping,
    TypeVar,
    Callable,
    Optional,
    Protocol,
    Sequence,
    cast,
)

import numpy as np
import plotly.colors  # type: ignore[import-untyped]
import plotly.graph_objects as go  # type: ignore[import-untyped]

from qm.results.simulator_samples import SimulatorSamples, SimulatorControllerSamples
from qm.type_hinting.simulator_types import (
    IqInfoType,
    PortToDucMap,
    ChirpInfoType,
    AdcAcquisitionType,
    PlayedWaveformType,
    PulserLocationType,
    WaveformReportType,
    PlayedAnalogWaveformType,
)


class HasPortsProtocol(Protocol):
    @property
    def ports(self) -> List[str]:
        raise NotImplementedError

    @property
    def controller(self) -> str:
        raise NotImplementedError

    @property
    def fem(self) -> int:
        raise NotImplementedError

    @property
    def element(self) -> str:
        raise NotImplementedError


T = TypeVar("T", bound="PlayedWaveform")


@dataclass(frozen=True)
class PlayedWaveform(metaclass=ABCMeta):
    waveform_name: str
    pulse_name: str
    length: int
    timestamp: int
    iq_info: IqInfoType
    element: str
    output_ports: List[int]
    controller: str
    pulser: Dict[str, Any]
    fem: int

    @staticmethod
    def build_initialization_dict(
        dict_description: PlayedWaveformType, formatted_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        attribute_dict: Dict[str, Any]
        if formatted_attribute_dict is None:
            attribute_dict = {}
        else:
            attribute_dict = deepcopy(formatted_attribute_dict)

        attribute_dict.update(
            pulse_name=dict_description["pulseName"],
            waveform_name=dict_description["waveformName"],
            timestamp=int(dict_description["timestamp"]),
            length=int(dict_description["length"]),
            iq_info=dict_description["iqInfo"],
            element=dict_description["quantumElements"],
            output_ports=[int(p) for p in dict_description["outputPorts"]],
            pulser=dict_description["pulser"],
            controller=dict_description["pulser"]["controllerName"],
            fem=int(dict_description["pulser"].get("femId", 0)) + 1,
        )
        return attribute_dict

    @classmethod
    def from_job_dict(cls: Type[T], dict_description: PlayedWaveformType) -> T:
        return cls(**cls.build_initialization_dict(dict_description))

    @property
    def ports(self) -> List[str]:
        return [str(p) for p in self.output_ports]

    @property
    def is_iq(self) -> bool:
        return self.iq_info["isPartOfIq"]

    @property
    def is_I_pulse(self) -> bool:
        return self.iq_info["isI"]

    @property
    def get_iq_association(self) -> str:
        if not self.is_iq:
            return ""
        return "I" if self.is_I_pulse else "Q"

    @property
    def ends_at(self) -> int:
        return self.timestamp + self.length

    @abstractmethod
    def to_string(self) -> str:
        return ""

    def __str__(self) -> str:
        return self.to_string()

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    def _common_attributes_to_printable_str_list(self) -> List[str]:
        waveform_type_string = "Type="
        if self.is_iq:
            waveform_type_string += f"IQ Type ({'I' if self.iq_info['isI'] else 'Q'})"
        else:
            waveform_type_string += "Single"
        return [
            f"Waveform Name={self.waveform_name}",
            f"Pulse name={remove_prefix(self.pulse_name, 'OriginPulseName=')}",
            f"Start Time={self.timestamp} ns",
            f"Length={self.length} ns",
            f"Element={self.element}",
            f"Output Ports={self.ports}",
            waveform_type_string,
        ]


# str.removeprefix exists only for python 3.9+, this is here for backward compatibility
def remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def format_float(f: float) -> str:
    return "{:.3f}".format(f)


def pretty_string_freq(f: float) -> str:
    if f < 1000:
        div, units = 1.0, "Hz"
    elif 1000 <= f < 1_000_000:
        div, units = 1000.0, "kHz"
    else:
        div, units = 1e6, "MHz"
    return f"{format_float(f / div).rstrip('0').rstrip('.')}{units}"


@dataclass(frozen=True)
class PlayedAnalogWaveform(PlayedWaveform):
    current_amp_elements: List[float]
    current_dc_offset_by_port: Dict[str, float]
    current_intermediate_frequency: float
    current_frame: List[float]
    current_correction_elements: List[float]
    chirp_info: Optional[ChirpInfoType]
    current_phase: float

    def to_custom_string(self, show_chirp: bool = True) -> str:
        _attributes = super()._common_attributes_to_printable_str_list()
        _attributes += (
            [
                f"{k}={v if self.is_iq else v[0]}"
                for k, v in [
                    ("Amplitude Values", [format_float(f) for f in self.current_amp_elements]),
                    ("Frame Values", [format_float(f) for f in self.current_frame]),
                    ("Correction Values", [format_float(f) for f in self.current_correction_elements]),
                ]
            ]
            + [
                f"Intermediate Frequency={pretty_string_freq(self.current_intermediate_frequency)}",
                f"Current DC Offset (By output ports)={ {k: format_float(v) for k, v in self.current_dc_offset_by_port.items()} }",
                f"Current Phase={format_float(self.current_phase)},",
            ]
            + ([] if (self.chirp_info is None or not show_chirp) else [f"chirp_info={self.chirp_info}"])
        )
        s = "AnalogWaveform(" + ("\n" + len("AnalogWaveform(") * " ").join(_attributes) + ")"
        return s

    def to_string(self) -> str:
        return self.to_custom_string()


def _fix_duc_idx(duc_idx: int) -> int:
    """This one is to overcome the returned 3 and 4 from the GW, when it's fixed, this one can be removed"""
    return (duc_idx - 1) % 2 + 1


@dataclass(frozen=True)
class PlayedMwAnalogWaveform(PlayedAnalogWaveform):
    port_to_duc: Dict[int, List[int]]

    @staticmethod
    def create_port_to_duc_mapping(port_to_duc: List[PortToDucMap]) -> Dict[int, List[int]]:
        return {int(p["port"]): [_fix_duc_idx(int(d)) for d in p["ducs"]] for p in port_to_duc}

    @property
    def ports(self) -> List[str]:
        to_return = []
        for port in self.output_ports:
            if port in self.port_to_duc:
                assert (
                    len(self.port_to_duc[port]) == 1
                ), f"Number of DUCs per port must equal 1, got {len(self.port_to_duc[port])}"
                to_return.append(f"{port}-{self.port_to_duc[port][0]}")
            else:
                to_return.append(str(port))
        return to_return


def _create_played_analog_waveform_from_job_dict(dict_description: PlayedWaveformType) -> PlayedAnalogWaveform:
    dict_description = cast(PlayedAnalogWaveformType, dict_description)
    pulse_chirp_info = dict_description["chirpInfo"]
    is_pulse_have_chirp = len(pulse_chirp_info["units"]) > 0 or len(pulse_chirp_info["rate"]) > 0
    formatted_attribute_list = dict(
        current_amp_elements=dict_description["currentGMatrixElements"],
        current_dc_offset_by_port=dict_description["currentDCOffsetByPort"],
        current_intermediate_frequency=dict_description["currentIntermediateFrequency"],
        current_frame=dict_description["currentFrame"],
        current_correction_elements=dict_description["currentCorrectionElements"],
        chirp_info=pulse_chirp_info if is_pulse_have_chirp else None,
        current_phase=dict_description.get("currentPhase", 0),
    )
    if dict_description.get("portToDuc") and any(d["ducs"] for d in dict_description["portToDuc"]):
        formatted_attribute_list["port_to_duc"] = PlayedMwAnalogWaveform.create_port_to_duc_mapping(
            dict_description["portToDuc"]
        )
        class_to_init: Type[PlayedAnalogWaveform] = PlayedMwAnalogWaveform
    else:
        class_to_init = PlayedAnalogWaveform
    initialization_dict = class_to_init.build_initialization_dict(dict_description, formatted_attribute_list)
    return class_to_init(**initialization_dict)


@dataclass(frozen=True)
class PlayedDigitalWaveform(PlayedWaveform):
    @classmethod
    def from_job_dict(cls: Type[T], dict_description: PlayedWaveformType) -> T:
        return cls(**cls.build_initialization_dict(dict_description))

    def to_string(self) -> str:
        s = (
            "DigitalWaveform("
            + ("\n" + len("DigitalWaveform(") * " ").join(self._common_attributes_to_printable_str_list())
            + ")"
        )
        return s


@dataclass(frozen=True)
class AdcAcquisition:
    start_time: int
    end_time: int
    process: str
    pulser: PulserLocationType
    quantum_element: str
    adc_ports: List[int]
    controller: str
    fem: int
    element: str

    @classmethod
    def from_job_dict(cls, dict_description: AdcAcquisitionType) -> "AdcAcquisition":
        return cls(
            start_time=int(dict_description["startTime"]),
            end_time=int(dict_description["endTime"]),
            process=dict_description["process"],
            pulser=dict_description["pulser"],
            quantum_element=dict_description["quantumElement"],
            adc_ports=[int(p) + 1 for p in dict_description["adc"]],
            controller=dict_description["pulser"]["controllerName"],
            fem=int(dict_description["pulser"].get("femId", 0)) + 1,
            element=dict_description["quantumElement"],
        )

    @property
    def ports(self) -> List[str]:
        return [str(p) for p in self.adc_ports]

    def to_string(self) -> str:
        return (
            "AdcAcquisition("
            + ("\n" + len("AdcAcquisition(") * " ").join(
                [
                    f"start_time={self.start_time}",
                    f"end_time={self.end_time}",
                    f"process={self.process}",
                    f"element={self.quantum_element}",
                    f"input_ports={self.adc_ports}",
                ]
            )
            + ")"
        )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclass
class FemToWaveformMap:
    analog_out: Dict[str, List[PlayedAnalogWaveform]]
    digital_out: Dict[str, List[PlayedDigitalWaveform]]
    analog_in: Dict[str, List[AdcAcquisition]]


class _SingleControllerMapping(Dict[int, FemToWaveformMap]):
    @property
    def num_analog_out_ports(self) -> int:
        return len(self.flat_analog_out)

    @property
    def num_digital_out_ports(self) -> int:
        return len(self.flat_digital_out)

    @property
    def num_analog_in_ports(self) -> int:
        return len(self.flat_analog_in)

    @property
    def flat_analog_out(self) -> Dict[str, List[PlayedAnalogWaveform]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.analog_out.items()}

    @property
    def flat_digital_out(self) -> Dict[str, List[PlayedDigitalWaveform]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.digital_out.items()}

    @property
    def flat_analog_in(self) -> Dict[str, List[AdcAcquisition]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.analog_in.items()}


@dataclass(frozen=True)
class WaveformReport:
    job_id: Union[int, str]
    analog_waveforms: List[PlayedAnalogWaveform]
    digital_waveforms: List[PlayedDigitalWaveform]
    adc_acquisitions: List[AdcAcquisition]

    @classmethod
    def from_dict(cls, d: WaveformReportType, job_id: Union[int, str] = -1) -> "WaveformReport":
        return cls(
            analog_waveforms=[_create_played_analog_waveform_from_job_dict(awf) for awf in d["analogWaveforms"]],
            digital_waveforms=[PlayedDigitalWaveform.from_job_dict(dwf) for dwf in d["digitalWaveforms"]],
            adc_acquisitions=[AdcAcquisition.from_job_dict(acq) for acq in d.get("adcAcquisitions", [])],
            job_id=job_id,
        )

    @property
    def waveforms(self) -> Sequence[PlayedWaveform]:
        return cast(List[PlayedWaveform], self.analog_waveforms) + cast(List[PlayedWaveform], self.digital_waveforms)

    @property
    def controllers_in_use(self) -> Sequence[str]:
        return sorted(self.fems_in_use_by_controller)

    @property
    def elements_in_report(self) -> Sequence[str]:
        wf_elements = {wf.element for wf in self.waveforms}
        adc_elements = {adc.element for adc in self.adc_acquisitions}
        return sorted(wf_elements | adc_elements)

    @property
    def fems_in_use_by_controller(self) -> Mapping[str, Sequence[int]]:
        fems_in_use = defaultdict(set)
        for ap in self.waveforms:
            fems_in_use[ap.controller].add(ap.fem)
        for adc in self.adc_acquisitions:
            fems_in_use[adc.controller].add(adc.fem)
        return {k: sorted(v) for k, v in fems_in_use.items()}

    @property
    def num_controllers_in_use(self) -> int:
        return len(self.controllers_in_use)

    def to_string(self) -> str:
        """
        Dumps the report into a (pretty-print) string.

        return: str
        """
        waveforms_str = [wf.to_string() for wf in self.waveforms]
        adc_string = [adc.to_string() for adc in self.adc_acquisitions]
        return "\n".join(waveforms_str + adc_string)

    def _transform_report_by_func(self, func: Callable[[HasPortsProtocol], bool]) -> "WaveformReport":
        return WaveformReport(
            analog_waveforms=list(filter(func, self.analog_waveforms)),
            digital_waveforms=list(filter(func, self.digital_waveforms)),
            adc_acquisitions=list(filter(func, self.adc_acquisitions)),
            job_id=self.job_id,
        )

    def report_by_controllers(self) -> Mapping[str, "WaveformReport"]:
        def create_filter_func(controller: str) -> Callable[[HasPortsProtocol], bool]:
            return lambda r: r.controller == controller

        by_controller_map: Dict[str, "WaveformReport"] = {}
        for con_name in self.controllers_in_use:
            con_filter = create_filter_func(con_name)
            by_controller_map[con_name] = self._transform_report_by_func(con_filter)

        return by_controller_map

    def report_by_elements(self) -> Mapping[str, "WaveformReport"]:
        def create_filter_func(_element: str) -> Callable[[HasPortsProtocol], bool]:
            return lambda r: r.element == _element

        by_element_map: Dict[str, "WaveformReport"] = {}
        for element in self.elements_in_report:
            element_filter = create_filter_func(element)
            by_element_map[element] = self._transform_report_by_func(element_filter)

        return by_element_map

    def report_by_controller_and_fems(self) -> Mapping[str, Mapping[int, "WaveformReport"]]:
        def create_filter_func(controller: str, _fem: int) -> Callable[[HasPortsProtocol], bool]:
            return lambda r: r.controller == controller and r.fem == _fem

        by_controller_fem_map: Dict[str, Dict[int, "WaveformReport"]] = {}
        for con_name in self.controllers_in_use:
            by_controller_fem_map[con_name] = {}
            for fem in self.fems_in_use_by_controller[con_name]:
                fem_filter = create_filter_func(con_name, fem)
                by_controller_fem_map[con_name][fem] = self._transform_report_by_func(fem_filter)

        return by_controller_fem_map

    def to_dict(self) -> Dict[str, Any]:
        """
        Dumps the report to a dictionary containing three keys:
            "analog_waveforms", "digital_waveforms", "acd_acquisitions".
        Each key holds the list of all the associate data.

        Returns:
            dict
        """
        return {
            "analog_waveforms": [awf.to_dict() for awf in self.analog_waveforms],
            "digital_waveforms": [dwf.to_dict() for dwf in self.digital_waveforms],
            "adc_acquisitions": [acq.to_dict() for acq in self.adc_acquisitions],
        }

    def _strict_get_report_by_output_ports(self) -> Mapping[str, _SingleControllerMapping]:
        report_by_controller_and_fem = self.report_by_controller_and_fems()
        result = {}
        for controller, fem_to_report_map in report_by_controller_and_fem.items():
            per_controller = {}
            for fem, report in fem_to_report_map.items():
                analog_out, digital_out, analog_in = defaultdict(list), defaultdict(list), defaultdict(list)
                for awf in report.analog_waveforms:
                    for p in awf.ports:
                        analog_out[p].append(awf)
                for dwf in report.digital_waveforms:
                    for p in dwf.ports:
                        digital_out[p].append(dwf)
                for adc in report.adc_acquisitions:
                    for p in adc.ports:
                        analog_in[p].append(adc)
                per_controller[fem] = FemToWaveformMap(
                    analog_out=dict(analog_out), digital_out=dict(digital_out), analog_in=dict(analog_in)
                )
            result[controller] = _SingleControllerMapping(per_controller)
        return result

    def get_report_by_output_ports(
        self, on_controller: Optional[str] = None
    ) -> Union[Mapping[str, _SingleControllerMapping], _SingleControllerMapping]:
        result = self._strict_get_report_by_output_ports()
        if on_controller is None:
            if self.num_controllers_in_use == 1:
                return result[self.controllers_in_use[0]]
            else:
                return result
        else:
            return result[on_controller]

    def create_plot(
        self,
        samples: Optional[SimulatorSamples] = None,
        controllers: Optional[Sequence[str]] = None,
        plot: bool = True,
        save_path: Optional[str] = None,
    ) -> None:
        """Creates a plot describing the pulses from each element to each port.
        See arguments description for further options.

        Args:
            samples: The raw samples as generated from the simulator. If not given, the plot will be generated without it.
            controllers: list of controllers to generate the plot. Each controller output will be saved as a different
                        file. If not given, take all the controllers who participate in the program.
            plot: Show the plot at the end of this function call.
            save_path: Save the plot to the given location. None for not saving.

        Returns:
            None
        """
        if controllers is None:
            controllers = self.controllers_in_use

        report_by_controllers = self.report_by_controllers()
        for con_name, report in report_by_controllers.items():
            if con_name not in controllers:
                continue

            if samples is None:
                con_builder = _WaveformPlotBuilder(report, self.job_id)
            else:
                con_builder = _WaveformPlotBuilderWithSamples(report, samples[con_name], self.job_id)

            if save_path is not None:
                dirname = os.path.dirname(save_path)
                filename = f"waveform_report_{con_name}_{self.job_id}"
                con_builder.save(dirname, filename)
            if plot:
                con_builder.plot()


@dataclass
class _MaxParallelTracesPerRow:
    analog: Dict[str, int]
    digital: Dict[str, int]


class _WaveformPlotBuilder:
    def __init__(self, wf_report: WaveformReport, job_id: Union[int, str] = -1):
        self._report = wf_report
        if wf_report.num_controllers_in_use > 1:
            raise RuntimeError(
                f"Plot Builder does not support plotting more than 1 controllers, yet. {os.linesep}"
                "Please provide a report containing a single controller."
            )
        self._job_id = job_id
        self._figure = go.Figure()
        self._already_registered_qe: Set[str] = set()
        self._colormap = self._get_qe_colorscale(wf_report.elements_in_report)
        report_by_output_ports = cast(_SingleControllerMapping, wf_report.get_report_by_output_ports())
        self._max_parallel_traces_per_row = self._calculate_max_parallel_traces_per_row(report_by_output_ports)
        self._num_rows = self._get_num_rows(report_by_output_ports)

        self._report_by_output_ports = report_by_output_ports
        self._setup_figure(self._num_rows)
        self._add_data()
        self._update_extra_features()

    @property
    def _samples_factor(self) -> int:
        return 1

    @staticmethod
    def _get_qe_colorscale(qe_in_use: Sequence[str]) -> Dict[str, str]:
        n_colors = len(qe_in_use)
        samples = plotly.colors.qualitative.Pastel + plotly.colors.qualitative.Safe
        if n_colors > len(samples):
            samples += plotly.colors.sample_colorscale(
                "turbo",
                [n / (n_colors - len(samples)) for n in range(n_colors - len(samples))],
            )
        return dict(zip(qe_in_use, samples))

    @staticmethod
    def _calculate_max_parallel_traces_per_row(
        report_by_output_port: _SingleControllerMapping,
    ) -> _MaxParallelTracesPerRow:
        def calc_row(_waveform_list: Sequence[PlayedWaveform]) -> int:
            max_in_row = 0
            functional_ts = sorted(
                [(r.timestamp, 1) for r in _waveform_list] + [(r.ends_at, -1) for r in _waveform_list],
                key=lambda t: t[0],
            )
            for _, f in functional_ts:
                max_in_row = max(max_in_row, max_in_row + f)
            return max_in_row

        analog_traces_per_row = {}
        digital_traces_per_row = {}
        for fem_idx, report in report_by_output_port.items():
            for output_port, waveform_list in report.analog_out.items():
                analog_traces_per_row[f"{fem_idx}-{output_port}"] = calc_row(waveform_list)
            for output_port, digital_waveform_list in report.digital_out.items():
                digital_traces_per_row[f"{fem_idx}-{output_port}"] = calc_row(digital_waveform_list)

        return _MaxParallelTracesPerRow(analog=analog_traces_per_row, digital=digital_traces_per_row)

    def _get_num_rows(self, report_by_output_ports: _SingleControllerMapping) -> int:
        num_rows = report_by_output_ports.num_analog_out_ports + report_by_output_ports.num_digital_out_ports
        num_rows *= self._samples_factor
        num_rows += report_by_output_ports.num_analog_in_ports
        return num_rows

    @property
    def _num_output_rows(self) -> int:
        return self._num_rows - self._report_by_output_ports.num_analog_in_ports

    @property
    def _num_analog_rows(self) -> int:
        return self._report_by_output_ports.num_analog_out_ports * self._samples_factor

    @property
    def _num_digital_rows(self) -> int:
        return self._report_by_output_ports.num_digital_out_ports * self._samples_factor

    def _is_row_analog(self, r: int) -> bool:
        return 1 <= r <= self._num_analog_rows

    def _is_row_digital(self, r: int) -> bool:
        return self._num_analog_rows < r <= self._num_output_rows

    def _is_row_analog_input(self, r: int) -> bool:
        return self._num_output_rows < r <= self._num_rows

    @property
    def _xrange(self) -> int:
        return max(self._report.waveforms, key=lambda x: x.ends_at).ends_at + 100

    @staticmethod
    def _is_intersect(r1: Tuple[int, int], r2: Tuple[int, int]) -> bool:
        return (r1[0] <= r2[0] <= r1[1]) or (r1[0] <= r2[1] <= r1[1]) or (r2[0] < r1[0] and r2[1] > r1[1])

    @staticmethod
    def _get_hover_text(played_waveform: PlayedWaveform) -> str:
        waveform_desc = played_waveform.to_string()
        if isinstance(played_waveform, PlayedAnalogWaveform):
            if played_waveform.chirp_info is not None:
                waveform_desc = played_waveform.to_custom_string(False)
                s = (
                    f"rate={played_waveform.chirp_info['rate']},units={played_waveform.chirp_info['units']},"
                    f" times={played_waveform.chirp_info['times']}\n"
                    + f"start_freq={pretty_string_freq(played_waveform.chirp_info['startFrequency'])}, "
                    + f"end_freq={pretty_string_freq(played_waveform.chirp_info['endFrequency'])}"
                )

                waveform_desc = f"<b>Chirp Pulse</b>\n({s})\n" + waveform_desc
        return "%{x}ns<br>" + waveform_desc.replace("\n", "</br>") + "<extra></extra>"

    def _get_output_port_waveform_plot_data(
        self, port_played_waveforms: Sequence[PlayedWaveform], x_axis_name: str, max_in_row: int
    ) -> List[go.Scatter]:
        graph_data: List[go.Scatter] = []
        levels: List[Tuple[int, int]] = []
        diff_between_traces, start_y = (0.2, 1.2) if max_in_row <= 7 else (1.4 / max_in_row, 1.45)
        y_level = [start_y] * 3
        for wf in port_played_waveforms:
            x_axis_points = (wf.timestamp, wf.ends_at)
            num_intersections = len([l for l in levels if self._is_intersect(l, x_axis_points)])
            levels.append(x_axis_points)
            prev_y = start_y if num_intersections == 0 else y_level[0]
            y_level = [prev_y - diff_between_traces] * 3
            graph_data.append(
                go.Scatter(
                    x=[x_axis_points[0], sum(x_axis_points) // 2, x_axis_points[1]],
                    y=y_level,
                    mode="lines+markers+text",
                    text=[
                        "",
                        f"{remove_prefix(wf.pulse_name, 'OriginPulseName=')}"
                        + (f"({wf.get_iq_association})" if wf.is_iq else ""),
                        "",
                    ],
                    hovertemplate=self._get_hover_text(wf),
                    textfont=dict(size=10),
                    xaxis=x_axis_name,
                    name=wf.element,
                    legendgroup=wf.element,
                    showlegend=not (wf.element in self._already_registered_qe),
                    marker=dict(
                        line=dict(width=2, color=self._colormap[wf.element]),
                        symbol=["line-ns", "line-ew", "line-ns"],
                    ),
                    line=dict(color=self._colormap[wf.element], width=5),
                )
            )
            self._already_registered_qe.add(wf.element)

        return graph_data

    def _add_plot_data_for_analog_output_port(
        self, figure_row_number: int, output_port: str, port_waveforms: Sequence[PlayedWaveform]
    ) -> None:
        self._add_plot_data_for_port(
            figure_row_number, self._max_parallel_traces_per_row.analog[output_port], port_waveforms
        )

    def _add_plot_data_for_port(
        self, figure_row_number: int, _max_parallel_traces_per_row: int, port_waveforms: Sequence[PlayedWaveform]
    ) -> None:
        if len(port_waveforms) == 0:
            return

        x_axis_name = self._get_x_axis_name(figure_row_number)
        port_wf_plot = self._get_output_port_waveform_plot_data(
            port_waveforms,
            x_axis_name,
            max_in_row=_max_parallel_traces_per_row,
        )
        row_number = figure_row_number * self._samples_factor
        self._figure.add_traces(port_wf_plot, rows=row_number, cols=1)

    @staticmethod
    def _get_x_axis_name(figure_row_number: int) -> str:
        return f"x{figure_row_number}"

    def _add_plot_data_for_digital_output_port(
        self, figure_row_number: int, output_port: str, port_waveforms: Sequence[PlayedWaveform]
    ) -> None:
        self._add_plot_data_for_port(
            figure_row_number, self._max_parallel_traces_per_row.digital[output_port], port_waveforms
        )

    def _add_plot_data_for_adc_port(
        self,
        figure_row_number: int,
        adc_port_acquisitions: List[AdcAcquisition],
    ) -> None:
        graph_data: List[go.Scatter] = []
        levels: List[Tuple[int, int]] = []
        y_level = [1.2] * 3
        for adc in adc_port_acquisitions:
            x_axis_points = (adc.start_time, adc.end_time)
            num_intersections = len([l for l in levels if self._is_intersect(l, x_axis_points)])
            levels.append(x_axis_points)
            prev_y = 1.2 if num_intersections == 0 else y_level[0]
            y_level = [prev_y - 0.2] * 3
            graph_data.append(
                go.Scatter(
                    x=[x_axis_points[0], sum(x_axis_points) // 2, x_axis_points[1]],
                    y=y_level,
                    mode="lines+markers+text",
                    text=["", f"{adc.process}", ""],
                    textfont=dict(size=10),
                    hovertemplate="%{x}ns<br>" + adc.to_string().replace("\n", "</br>") + "<extra></extra>",
                    name=adc.quantum_element,
                    legendgroup=adc.quantum_element,
                    showlegend=not (adc.quantum_element in self._already_registered_qe),
                    marker=dict(
                        line=dict(width=2, color=self._colormap[adc.quantum_element]),
                        symbol=["line-ns", "line-ew", "line-ns"],
                    ),
                    line=dict(color=self._colormap[adc.quantum_element], width=5),
                )
            )
            self._already_registered_qe.add(adc.quantum_element)

        self._figure.add_traces(graph_data, rows=figure_row_number, cols=1)

    def _add_data(self) -> None:
        for figure_row_number, (output_port, port_waveforms_list) in enumerate(
            self._report_by_output_ports.flat_analog_out.items()
        ):
            self._add_plot_data_for_analog_output_port(figure_row_number + 1, output_port, port_waveforms_list)

        for (figure_row_number, (output_port, digital_port_waveforms_list)) in enumerate(
            self._report_by_output_ports.flat_digital_out.items()
        ):
            self._add_plot_data_for_digital_output_port(
                figure_row_number + self._report_by_output_ports.num_analog_out_ports + 1,
                output_port,
                digital_port_waveforms_list,
            )

        for figure_row_number, adc_acquisition_list in enumerate(self._report_by_output_ports.flat_analog_in.values()):
            self._add_plot_data_for_adc_port(figure_row_number + self._num_output_rows + 1, adc_acquisition_list)

    def _update_extra_features(self) -> None:
        all_x_axis_names = sorted(
            [a for a in self._figure.layout.__dir__() if a.startswith("xaxis")],
            key=lambda s: int(s.removeprefix("xaxis")) if s.removeprefix("xaxis").isnumeric() else 0,
        )
        all_xaxis_names_short = {
            k: "x" + k.removeprefix("xaxis") if k.removeprefix("xaxis").isnumeric() else "" for k in all_x_axis_names
        }
        bottommost_x_axis = all_x_axis_names[-1]
        self._figure.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="left",
                    active=0,
                    buttons=list(
                        [
                            dict(
                                args=[
                                    {k + ".matches": all_xaxis_names_short[bottommost_x_axis] for k in all_x_axis_names}
                                ],
                                label="Shared",
                                method="relayout",
                            ),
                            dict(
                                args=[{k + ".matches": v for k, v in all_xaxis_names_short.items()}],
                                label="Distinct",
                                method="relayout",
                            ),
                        ]
                    ),
                    showactive=True,
                    x=1,
                    xanchor="right",
                    y=1,
                    yanchor="bottom",
                    font=dict(size=10),
                ),
            ]
        )
        self._figure.add_annotation(
            dict(
                text="X-Axis scrolling method:",
                showarrow=False,
                x=1,
                y=1,
                yref="paper",
                yshift=40,
                yanchor="bottom",
                xref="paper",
                align="left",
            )
        )
        self._figure.update_layout(
            modebar_remove=[
                "autoscale",
                "autoscale2d",
                "lasso",
            ]
        )

        source_path = os.path.join(os.path.dirname(__file__), "sources", "logo_qm_square.png")

        im = base64.b64encode(open(source_path, "rb").read())
        self._figure.add_layout_image(
            source="data:image/png;base64,{}".format(im.decode()),
            xref="paper",
            yref="paper",
            x=0,
            y=1,
            sizex=0.1,
            sizey=0.1,
            xanchor="center",
            yanchor="bottom",
        )

    @property
    def _subplot_titles(self) -> Sequence[Union[str, Sequence[str]]]:
        titles = (
            [f"Analog-Out-{a}" for a in self._report_by_output_ports.flat_analog_out]
            + [f"Digital-Out-{d}" for d in self._report_by_output_ports.flat_digital_out]
            + [f"Analog-In-{ai}" for ai in self._report_by_output_ports.flat_analog_in]
        )
        return titles

    def _get_subplot_specs(self, num_rows: int) -> List[List[Dict[str, float]]]:
        return [[{"t": 1 / (num_rows * 4)}]] * num_rows

    def _setup_figure(self, num_rows: int, minimum_number_of_rows: int = 4) -> None:
        num_rows = max(num_rows, minimum_number_of_rows)
        self._figure.set_subplots(
            rows=num_rows,
            cols=1,
            subplot_titles=self._subplot_titles,
            vertical_spacing=0.1 / num_rows,
            specs=self._get_subplot_specs(num_rows),
        )

        self._figure.update_layout(
            hovermode="closest",
            hoverdistance=5,
            height=160 * num_rows,
            title=dict(
                text=(
                    f"Waveform Report (connection: {self._report.controllers_in_use[0]})"
                    + (" for job: {}".format(self._job_id) if self._job_id != -1 else "")
                ),
                x=0.5,
                xanchor="center",
                yanchor="auto",
                xref="paper",
            ),
            legend=dict(title="Elements", y=0.98, yanchor="top"),
        )
        self._figure.add_annotation(
            dict(
                text=f"Created at {datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}",
                showarrow=False,
                x=0.5,
                y=1,
                yref="paper",
                yshift=20,
                yanchor="bottom",
                xref="paper",
                xanchor="center",
            )
        )

        for idx in range(1, num_rows + 1):
            self._figure.update_xaxes(range=[0, self._xrange], row=idx, col=1)

        self._update_axes_for_samples()

        inc = self._samples_factor
        input_rows_first_idx = self._num_output_rows + 1
        output_rows_indices = list(range(inc, input_rows_first_idx, inc))
        input_rows_indices = list(
            range(input_rows_first_idx, input_rows_first_idx + self._report_by_output_ports.num_analog_in_ports)
        )
        for idx in output_rows_indices + input_rows_indices:
            self._figure.update_yaxes(
                range=[-0.5, 1.5],
                showticklabels=False,
                tickvals=[-0.5],
                ticklen=50,
                tickcolor="#000000",
                showgrid=False,
                zeroline=False,
                row=idx,
                col=1,
            )
            self._figure.update_xaxes(title=dict(text="Time(ns)", standoff=5, font=dict(size=9)), row=idx, col=1)

    def _update_axes_for_samples(self) -> None:
        return

    def plot(self) -> None:
        self._figure.show(renderer="browser")

    def save(self, basedir: str = "", filename: str = "") -> None:
        os.makedirs(basedir, exist_ok=True)
        if filename == "":
            filename = f"waveform_report_{self._job_id}"
        if not os.path.splitext(filename)[1] == "html":
            filename += ".html"

        path = os.path.join(basedir, filename)
        with open(path, "w", encoding="UTF-8") as f:
            self._figure.write_html(f)


class _WaveformPlotBuilderWithSamples(_WaveformPlotBuilder):
    def __init__(self, wf_report: WaveformReport, samples: SimulatorControllerSamples, job_id: Union[int, str] = -1):
        self._samples = samples
        super().__init__(wf_report, job_id)

    @property
    def _samples_factor(self) -> int:
        return 2

    @property
    def _xrange(self) -> int:
        first_key = next(iter(self._samples.analog.keys()))
        return len(self._samples.analog[first_key])

    def _add_plot_data_for_analog_output_port(
        self, figure_row_number: int, output_port: str, port_waveforms: Sequence[PlayedWaveform]
    ) -> None:
        if len(port_waveforms) == 0:
            return

        self._add_plot_data_for_port(
            figure_row_number, self._max_parallel_traces_per_row.analog[output_port], port_waveforms
        )

        port_samples = self._samples.analog[output_port]
        sampling_rate = self._samples.analog_sampling_rate[output_port] / 1e9
        t = list(x / sampling_rate for x in range(len(port_samples)))
        self._add_trace_to_figure(figure_row_number, t, np.real(port_samples).tolist())
        if isinstance(port_samples[0], complex):
            self._add_trace_to_figure(figure_row_number, t, np.imag(port_samples).tolist())

    def _add_plot_data_for_digital_output_port(
        self, figure_row_number: int, output_port: str, port_waveforms: Sequence[PlayedWaveform]
    ) -> None:
        if len(port_waveforms) == 0:
            return

        self._add_plot_data_for_port(
            figure_row_number, self._max_parallel_traces_per_row.digital[output_port], port_waveforms
        )

        port_samples = self._fetch_digital_samples(output_port)
        t = list(range(len(port_samples)))  # here we assume 1ns sampling rate
        self._add_trace_to_figure(figure_row_number, t, port_samples)

    def _fetch_digital_samples(self, output_port: Union[int, str]) -> Sequence[int]:
        port_samples = self._samples.digital.get(str(output_port))
        if port_samples is None:
            logging.log(logging.WARNING, f"Could not find digital samples for output port {output_port}")
            return [0] * self._xrange
        return [int(_x) for _x in port_samples]

    def _add_trace_to_figure(
        self, figure_row_number: int, t: Sequence[float], port_samples: Union[Sequence[int], Sequence[float]]
    ) -> None:
        self._figure.add_trace(
            go.Scatter(
                x=t,
                y=port_samples,
                showlegend=False,
                xaxis=self._get_x_axis_name(figure_row_number),
                hovertemplate="%{x}ns, %{y}v<extra></extra>",
            ),
            row=figure_row_number * 2 - 1,
            col=1,
        )

    @property
    def _subplot_titles(self) -> Sequence[Union[str, Sequence[str]]]:
        _titles = [f"Analog-Out-{a}" for a in self._report_by_output_ports.flat_analog_out] + [
            f"Digital-Out-{d}" for d in self._report_by_output_ports.flat_digital_out
        ]
        zipped: Sequence[Tuple[str, Sequence[str]]] = list(zip(_titles, [[]] * len(_titles)))
        titles: Sequence[Union[str, Sequence[str]]] = [item for z in zipped for item in z] + [
            f"Analog-In-{a}" for a in self._report_by_output_ports.flat_analog_in
        ]
        return titles

    def _get_subplot_specs(self, num_rows: int) -> List[List[Dict[str, float]]]:
        specs = ([[{"t": 1.2 / (num_rows * 5)}]] + [[{"b": 1.2 / (num_rows * 5)}]]) * (self._num_output_rows // 2) + [
            [{"t": 1 / (num_rows * 4)}]
        ] * self._report_by_output_ports.num_analog_in_ports
        if len(specs) < num_rows:
            specs += [[{}]] * (num_rows - len(specs))
        return specs

    def _update_axes_for_samples(self) -> None:
        for r in range(1, self._num_output_rows, 2):
            sample_y_range = [-0.1, 1.1] if self._is_row_digital(r) else [-0.6, 0.6]
            self._figure.update_yaxes(range=sample_y_range, row=r, col=1)
            self._figure.update_xaxes(showticklabels=False, row=r, col=1)
            self._figure.update_yaxes(title=dict(text="Voltage(v)", standoff=5, font=dict(size=9)), row=r, col=1)



import logging
from typing import TYPE_CHECKING, Tuple, Generic, TypeVar, Optional

from qm.persistence import BaseStore
from qm.program.program import Program
from qm.api.v2.job_api.job_api import JobApi
from qm.jobs.pending_job import QmPendingJob
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.compiler import CompilerOptionArguments

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class JobNotFoundError(Exception):
    pass


JobTypeVar = TypeVar("JobTypeVar", JobApi, QmPendingJob)


class QmNewApiNotImplementedError(NotImplementedError):
    pass


class QmQueueBase(Generic[JobTypeVar]):
    def __init__(self, store: BaseStore, capabilities: ServerCapabilities):
        self._store = store
        self._capabilities = capabilities

    def _get_pending_jobs(
        self,
        job_id: Optional[str] = None,
        position: Optional[int] = None,
        user_id: Optional[str] = None,
    ) -> Tuple[JobTypeVar, ...]:
        raise QmNewApiNotImplementedError

    def add(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> JobTypeVar:
        raise QmNewApiNotImplementedError

    def add_compiled(self, program_id: str) -> JobTypeVar:
        raise QmNewApiNotImplementedError

    def add_to_start(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> JobTypeVar:
        """Adds a QMJob to the start of the queue.
        Programs in the queue will play as soon as possible.

        Args:
            program: A QUA program
            compiler_options: Optional arguments for compilation

        """
        raise QmNewApiNotImplementedError

    @property
    def count(self) -> int:
        """Get the number of jobs currently on the queue

        Returns:
            The number of jobs in the queue

        Example:
            ```python
            qm.queue.count
            ```
        """
        return len(self._get_pending_jobs())

    def __len__(self) -> int:
        return self.count

    @property
    def pending_jobs(self) -> Tuple[JobTypeVar, ...]:
        """Returns all currently pending jobs

        Returns:
            A list of all the currently pending jobs
        """
        return self._get_pending_jobs()

    def get(self, job_id: str) -> JobTypeVar:
        """Get a pending job object by job_id

        Args:
            job_id: a QMJob id

        Returns:
            The pending job

        Example:
            ```python
            qm.queue.get(job_id)
            ```
        """
        jobs = self._get_pending_jobs(job_id)
        if len(jobs) == 0:
            raise JobNotFoundError()
        return jobs[0]

    def get_at(self, position: int) -> JobTypeVar:
        raise QmNewApiNotImplementedError

    def get_by_user_id(self, user_id: str) -> Tuple[JobTypeVar, ...]:
        return self._get_pending_jobs(user_id=user_id)

    def remove_by_id(self, job_id: str) -> int:
        raise QmNewApiNotImplementedError

    def remove_by_position(self, position: int) -> int:
        raise QmNewApiNotImplementedError

    def remove_by_user_id(self, user_id: str) -> int:
        raise QmNewApiNotImplementedError

    def __getitem__(self, position: int) -> JobTypeVar:
        return self.get_at(position)

    def clear(self) -> int:
        raise QmNewApiNotImplementedError



import re
import logging
from typing import Dict, Tuple

import httpx
from betterproto.lib.google.protobuf import Empty

from qm.api.models.server_details import ResponseConnectionDetails
from qm.exceptions import QmRedirectionError, QmLocationParsingError

logger = logging.getLogger(__name__)


def _parse_location(location_header: str) -> Tuple[str, int]:
    match = re.match("(?P<host>[^:]*):(?P<port>[0-9]*)(/(?P<url>.*))?", location_header)
    if match is None:
        raise QmLocationParsingError(f"Could not parse new host and port (location header: {location_header})")
    host, port, _, __ = match.groups()
    if not port.isdigit() or not host:
        raise QmLocationParsingError(f"Could not parse new port (location header: {location_header})")
    return str(host), int(port)


def parse_octaves(raw_response: str) -> Dict[str, Tuple[str, int]]:
    octaves = {}
    for octave_details in raw_response.split(";"):
        if octave_details:
            name_and_location = octave_details.split(",")
            if len(name_and_location) != 2:
                raise QmLocationParsingError(
                    f"Could not parse octave name and location from '{octave_details}' (raw response: {raw_response})"
                )
            octaves[name_and_location[0]] = _parse_location(name_and_location[1])
    return octaves


async def send_redirection_check(
    host: str, port: int, headers: Dict[str, str], timeout: float
) -> ResponseConnectionDetails:
    extended_headers = {"content-type": "application/grpc", "te": "trailers", **headers}
    async with httpx.AsyncClient(http2=True, follow_redirects=False, http1=False, timeout=timeout) as client:
        response = await client.post(f"http://{host}:{port}", headers=extended_headers, content=bytes(Empty()))
    if response.status_code == 400:
        if headers.get("any_cluster", "false") == "false":
            cluster_name = f"cluster '{headers['cluster_name']}'"
        else:
            cluster_name = "any cluster"
        raise QmRedirectionError(f"Connected to server at in {host}:{port}. Could not find {cluster_name}.")
    if response.status_code != 302:
        return ResponseConnectionDetails(host, port, {})

    new_host, new_port = _parse_location(response.headers["location"])
    octaves = parse_octaves(response.headers.get("octaves", ""))

    return ResponseConnectionDetails(new_host, new_port, octaves)



import json
import asyncio
import logging
import functools
from abc import ABCMeta, abstractmethod
from typing_extensions import ParamSpec
from typing import Type, Generic, TypeVar, Callable, Optional

import grpclib.exceptions
from grpclib import Status
from grpclib.client import Channel
from grpclib.config import Configuration
from betterproto.grpc.grpclib_client import ServiceStub
from grpclib.events import SendMessage, SendRequest, RecvInitialMetadata, listen

from qm.utils.async_utils import run_async
from qm.utils.general_utils import is_debug
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMTimeoutError, QMConnectionError

logger = logging.getLogger(__name__)

Ret = TypeVar("Ret")
P = ParamSpec("P")


class GatewayNotImplementedError(NotImplementedError):
    pass


def connection_error_handle_decorator(func: Callable[P, Ret]) -> Callable[P, Ret]:
    @functools.wraps(func)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> Ret:
        try:
            return func(*args, **kwargs)
        except grpclib.exceptions.GRPCError as e:
            if is_debug():
                logger.exception("Encountered connection error from QOP")
            if e.status == Status.UNIMPLEMENTED:
                raise GatewayNotImplementedError(
                    f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
                ) from e
            raise QMConnectionError(
                f"Encountered connection error from QOP: details: {e.message}, status: {e.status}"
            ) from e
        except asyncio.TimeoutError as e:
            if is_debug():
                logger.exception(f"Timeout reached while running '{func.__name__}'")
            raise QMTimeoutError(f"Timeout reached while running '{func.__name__}'") from e

    return wrapped


T = TypeVar("T")
StubType = TypeVar("StubType", bound=ServiceStub)


def connection_error_handle() -> Callable[[Type[T]], Type[T]]:
    def decorate(cls: Type[T]) -> Type[T]:
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, connection_error_handle_decorator(getattr(cls, attr)))
        return cls

    return decorate


async def create_channel(connection_details: ConnectionDetails) -> Channel:
    return Channel(
        host=connection_details.host,
        port=connection_details.port,
        ssl=connection_details.ssl_context,
        config=Configuration(
            http2_connection_window_size=connection_details.max_message_size,
            http2_stream_window_size=connection_details.max_message_size,
        ),
    )


class BaseApi(Generic[StubType], metaclass=ABCMeta):
    def __init__(self, connection_details: ConnectionDetails):
        self._connection_details = connection_details

        self._channel = run_async(create_channel(self._connection_details))
        self._stub: StubType = self._stub_class(self._channel)

        if self._connection_details.debug_data:
            self._create_debug_data_event()

        self._create_add_headers_event()
        self._timeout: Optional[float] = self._connection_details.timeout

    @property
    def connection_details(self) -> ConnectionDetails:
        return self._connection_details

    @property
    @abstractmethod
    def _stub_class(self) -> Type[StubType]:
        pass

    def _create_debug_data_event(self) -> None:
        async def intercept_response(event: RecvInitialMetadata) -> None:
            assert self._connection_details.debug_data is not None

            metadata = event.metadata
            logger.debug(f"Collected response metadata: {json.dumps(dict(metadata), indent=4)}")
            self._connection_details.debug_data.append(metadata)

        async def send_request_debug(event: SendRequest) -> None:
            logger.debug("-----------request start-----------")
            logger.debug("   ---    request headers    ---   ")
            logger.debug(f"method:       {event.method_name}")
            logger.debug(f"metadata:     {json.dumps(dict(event.metadata), indent=4)}")
            logger.debug(f"content type: {event.content_type}")
            if event.deadline:
                deadline = event.deadline.time_remaining()
            else:
                deadline = None
            logger.debug(f"deadline:     {deadline}")

        async def send_message_debug(event: SendMessage) -> None:
            logger.debug("   ---    request message    ---   ")
            try:
                logger.debug(f"message:      {event.message.to_json(4)}")
            except TypeError:
                pass
            logger.debug("------------end request------------")

        listen(self._channel, RecvInitialMetadata, intercept_response)
        listen(self._channel, SendRequest, send_request_debug)
        listen(self._channel, SendMessage, send_message_debug)

    def _create_add_headers_event(self) -> None:
        async def add_headers(event: SendRequest) -> None:
            event.metadata.update(self._connection_details.headers)

        listen(self._channel, SendRequest, add_headers)

    @property
    def channel(self) -> Channel:
        return self._channel

    def __del__(self) -> None:
        self.channel.close()



from qm.api.v2.job_api.element_api import JobElement
from qm.api.v2.job_api.job_api import JobApi, JobStatus
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi

__all__ = [
    "JobApi",
    "JobStatus",
    "JobElement",
]



import os
import json
import uuid
import warnings
from dataclasses import field, asdict, dataclass
from typing import Any, Dict, Union, Optional, TypedDict, cast

from qm.type_hinting.general import PathLike

CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".qm", "config.json")


class UserConfigType(TypedDict, total=False):
    quantumMachinesManager_port: Optional[int]
    quantumMachinesManager_host: str
    quantumMachinesManager_strict_healthcheck: bool
    quantumMachinesManager_user_token: str
    quantumMachinesManager_managerPort: Optional[int]
    logging_level: Union[int, str]
    upload_logs: bool
    datadog_token: str
    organization: str
    octave_calibration_db_path: Optional[str]


@dataclass
class UserConfig:
    SESSION_ID = uuid.uuid4()
    quantumMachinesManager_port: Optional[int] = field(default=None)
    quantumMachinesManager_host: Optional[str] = field(default=None)
    quantumMachinesManager_strict_healthcheck: bool = field(default=True)
    quantumMachinesManager_user_token: str = field(default="")
    quantumMachinesManager_managerPort: Optional[int] = field(default=None)
    # This field is not is use anymore
    logging_level: Union[int, str] = field(default="INFO")
    upload_logs: bool = field(default=False)
    datadog_token: str = field(default="")
    organization: str = field(default="Unknown")
    octave_calibration_db_path: Optional[str] = field(default=None)

    @property
    def manager_port(self) -> Optional[int]:
        return self.quantumMachinesManager_port

    @property
    def manager_host(self) -> Optional[str]:
        return self.quantumMachinesManager_host

    @property
    def strict_healthcheck(self) -> bool:
        return self.quantumMachinesManager_strict_healthcheck

    @property
    def user_token(self) -> str:
        return self.quantumMachinesManager_user_token

    @classmethod
    def create_from_file(cls) -> "UserConfig":
        raw_dict = cls._read_json(CONFIG_PATH)
        return cls(**raw_dict)

    def write_to_file(self) -> None:
        config_dict: UserConfigType = cast(UserConfigType, asdict(self))
        self._write_json(config_dict, CONFIG_PATH)

    @staticmethod
    def _read_json(file_path: PathLike) -> UserConfigType:
        if os.path.exists(file_path):
            with open(file_path, "r") as file:
                return cast(UserConfigType, json.load(file))
        else:
            return {}

    @staticmethod
    def _write_json(config: UserConfigType, file_path: str) -> None:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as file:
            return json.dump(config, file, indent=4)

    @property
    def enable_user_stdout(self) -> bool:
        return os.environ.get("QM_DISABLE_STREAMOUTPUT", None) is None

    @property
    def default_logging_format(self) -> str:
        return "%(asctime)s - qm - %(levelname)-8s - %(message)s"

    @property
    def datadog_handler_config(self) -> Dict[str, Any]:
        warnings.warn("This property is deprecated, and will be removed in the next version", DeprecationWarning)
        return self._datadog_handler_config

    @property
    def _datadog_handler_config(self) -> Dict[str, Any]:
        assert self.user_token, "No user token is defined"
        return {
            "class": "qm.datadog_api.DatadogHandler",
            "user_id": self.user_token,
            "organization": self.organization,
            "user_token": self.datadog_token,
            "session_id": self.SESSION_ID,
        }

    @property
    def logging_config_dict(self) -> Dict[str, Any]:
        warnings.warn("This property is deprecated, and will be removed in the next version", DeprecationWarning)
        default_formatter = "default"
        formatters = {default_formatter: {"format": self.default_logging_format}}
        handlers = {}
        if self.enable_user_stdout:
            handlers["stdout"] = {
                "class": "logging.StreamHandler",
                "formatter": default_formatter,
                "stream": "ext://sys.stdout",
            }
        if self.upload_logs and self.datadog_token:
            handlers["datadog"] = self._datadog_handler_config
        return {
            "version": 1,
            "formatters": formatters,
            "handlers": handlers,
            "loggers": {"qm": {"level": self.logging_level, "handlers": list(handlers)}},
        }


def load_user_config() -> UserConfig:
    return UserConfig.create_from_file()


def _generate_random_id() -> str:
    return str(uuid.uuid4()).split("-")[-1]


def update_old_user_config(
    old_config: UserConfig,
    host: Optional[str] = None,
    port: Optional[int] = None,
    send_anonymous_logs: Optional[bool] = None,
    datadog_token: str = "",
    organization: str = "Unknown",
) -> UserConfig:
    user_token = old_config.user_token or _generate_random_id()
    upload_logs = old_config.upload_logs if send_anonymous_logs is None else send_anonymous_logs
    return UserConfig(
        upload_logs=upload_logs,
        quantumMachinesManager_user_token=user_token,
        datadog_token=datadog_token,
        organization=organization,
        quantumMachinesManager_host=host,
        quantumMachinesManager_port=port,
    )


def create_new_user_config(
    host: Optional[str] = None,
    port: Optional[int] = None,
    send_anonymous_logs: Optional[bool] = None,
    datadog_token: str = "",
    organization: str = "Unknown",
) -> None:
    """Creates a user config file (running it requires writing permission). Setting
    a host and a port removes the need to specify them each time opening a QuantumMachinesManager.
    All parameters are optional.

    Args:
        host: ip of the host (string)
        port: port number (int)
        send_anonymous_logs: (bool) Setting this parameter to `True`
            means you agree to send anonymously the logs that the qm
            module produces.
        datadog_token: string of token to send logs.
        organization: Name of your organization.
    """
    old_config = UserConfig.create_from_file()
    config = update_old_user_config(
        old_config,
        host=host,
        port=port,
        send_anonymous_logs=send_anonymous_logs,
        datadog_token=datadog_token,
        organization=organization,
    )
    config.write_to_file()
